from typing import Dict, Optional, Tuple, Union

import numpy as np
from ConfigSpace.configuration_space import ConfigurationSpace
from ConfigSpace.hyperparameters import UniformFloatHyperparameter

import autosklearn.pipeline.implementations.MinorityCoalescer
from autosklearn.askl_typing import FEAT_TYPE_TYPE
from autosklearn.pipeline.base import DATASET_PROPERTIES_TYPE, PIPELINE_DATA_DTYPE
from autosklearn.pipeline.components.base import AutoSklearnPreprocessingAlgorithm
from autosklearn.pipeline.constants import DENSE, INPUT, SPARSE, UNSIGNED_DATA


class MinorityCoalescer(AutoSklearnPreprocessingAlgorithm):
    """Group categories whose occurence is less than a specified minimum fraction."""

    def __init__(
        self,
        feat_type: Optional[FEAT_TYPE_TYPE] = None,
        minimum_fraction: float = 0.01,
        random_state: Optional[Union[int, np.random.RandomState]] = None,
    ) -> None:
        self.minimum_fraction = minimum_fraction

    def fit(
        self, X: PIPELINE_DATA_DTYPE, y: Optional[PIPELINE_DATA_DTYPE] = None
    ) -> "MinorityCoalescer":
        self.minimum_fraction = float(self.minimum_fraction)

        self.preprocessor = (
            autosklearn.pipeline.implementations.MinorityCoalescer.MinorityCoalescer(
                minimum_fraction=self.minimum_fraction
            )
        )
        self.preprocessor.fit(X, y)
        return self

    def transform(self, X: PIPELINE_DATA_DTYPE) -> PIPELINE_DATA_DTYPE:
        if self.preprocessor is None:
            raise NotImplementedError()
        return self.preprocessor.transform(X)

    @staticmethod
    def get_properties(
        dataset_properties: Optional[DATASET_PROPERTIES_TYPE] = None,
    ) -> Dict[str, Optional[Union[str, int, bool, Tuple]]]:
        return {
            "shortname": "coalescer",
            "name": "Categorical minority coalescer",
            "handles_regression": True,
            "handles_classification": True,
            "handles_multiclass": True,
            "handles_multilabel": True,
            "handles_multioutput": True,
            # TODO find out of this is right!
            "handles_sparse": True,
            "handles_dense": True,
            "input": (DENSE, SPARSE, UNSIGNED_DATA),
            "output": (INPUT,),
        }

    @staticmethod
    def get_hyperparameter_search_space(
        feat_type: Optional[FEAT_TYPE_TYPE] = None,
        dataset_properties: Optional[DATASET_PROPERTIES_TYPE] = None,
    ) -> ConfigurationSpace:
        cs = ConfigurationSpace()
        minimum_fraction = UniformFloatHyperparameter(
            "minimum_fraction", lower=0.0001, upper=0.5, default_value=0.01, log=True
        )
        cs.add_hyperparameter(minimum_fraction)
        return cs



from typing import Dict, Optional, Tuple, Union

import numpy as np
from ConfigSpace.configuration_space import ConfigurationSpace
from ConfigSpace.hyperparameters import UniformFloatHyperparameter
from scipy import sparse
from sklearn.exceptions import NotFittedError

from autosklearn.askl_typing import FEAT_TYPE_TYPE
from autosklearn.pipeline.base import DATASET_PROPERTIES_TYPE, PIPELINE_DATA_DTYPE
from autosklearn.pipeline.components.base import AutoSklearnPreprocessingAlgorithm
from autosklearn.pipeline.components.data_preprocessing.rescaling.abstract_rescaling import (  # noqa: E501
    Rescaling,
)
from autosklearn.pipeline.constants import (
    DENSE,
    INPUT,
    SIGNED_DATA,
    SPARSE,
    UNSIGNED_DATA,
)


class RobustScalerComponent(Rescaling, AutoSklearnPreprocessingAlgorithm):
    def __init__(
        self,
        q_min: float,
        q_max: float,
        random_state: Optional[Union[int, np.random.RandomState]] = None,
    ) -> None:
        from sklearn.preprocessing import RobustScaler

        self.q_min = q_min
        self.q_max = q_max
        self.preprocessor = RobustScaler(
            quantile_range=(self.q_min, self.q_max),
            copy=False,
        )

    @staticmethod
    def get_properties(
        dataset_properties: Optional[DATASET_PROPERTIES_TYPE] = None,
    ) -> Dict[str, Optional[Union[str, int, bool, Tuple]]]:
        return {
            "shortname": "RobustScaler",
            "name": "RobustScaler",
            "handles_regression": True,
            "handles_classification": True,
            "handles_multiclass": True,
            "handles_multilabel": True,
            "handles_multioutput": True,
            "is_deterministic": True,
            # TODO find out if this is right!
            "handles_sparse": True,
            "handles_dense": True,
            "input": (SPARSE, DENSE, UNSIGNED_DATA),
            "output": (INPUT, SIGNED_DATA),
            "preferred_dtype": None,
        }

    @staticmethod
    def get_hyperparameter_search_space(
        feat_type: Optional[FEAT_TYPE_TYPE] = None,
        dataset_properties: Optional[DATASET_PROPERTIES_TYPE] = None,
    ) -> ConfigurationSpace:
        cs = ConfigurationSpace()
        q_min = UniformFloatHyperparameter("q_min", 0.001, 0.3, default_value=0.25)
        q_max = UniformFloatHyperparameter("q_max", 0.7, 0.999, default_value=0.75)
        cs.add_hyperparameters((q_min, q_max))
        return cs

    def fit(
        self, X: PIPELINE_DATA_DTYPE, y: Optional[PIPELINE_DATA_DTYPE] = None
    ) -> "AutoSklearnPreprocessingAlgorithm":
        if self.preprocessor is None:
            raise NotFittedError()
        if sparse.isspmatrix(X):
            self.preprocessor.set_params(with_centering=False)

        return super(RobustScalerComponent, self).fit(X, y)



import typing

import copy
import itertools

import numpy as np
import pandas as pd
import scipy.stats
import sklearn.ensemble


class AbstractSelector:
    def fit(
        self,
        X: pd.DataFrame,
        y: pd.DataFrame,
        minima: typing.Dict[int, typing.Dict[str, float]],
        maxima: typing.Dict[int, typing.Dict[str, float]],
    ) -> None:
        raise NotImplementedError()

    def predict(
        self, X: pd.DataFrame, y: typing.Optional[pd.DataFrame] = None
    ) -> pd.DataFrame:
        prediction = self._predict(X, y)
        for col, series in prediction.iteritems():
            assert series.dtype == float, (col, series)
        np.testing.assert_array_almost_equal(
            prediction.sum(axis="columns").to_numpy(),
            np.ones(X.shape[0]),
            err_msg=prediction.to_csv(),
        )
        return prediction

    def _predict(
        self, X: pd.DataFrame, y: typing.Optional[pd.DataFrame]
    ) -> pd.DataFrame:
        raise NotImplementedError()


class OneVSOneSelector(AbstractSelector):
    def __init__(self, configuration, random_state, tie_break_order):
        self.configuration = configuration
        self.single_strategy_idx = None
        self.rng = random_state
        self.tie_break_order = tie_break_order
        self.models = None
        self.target_indices = None
        self.selectors_ = None
        self.weights_ = {}
        self.strategies_ = None

    def fit(
        self,
        X: pd.DataFrame,
        y: pd.DataFrame,
        minima: typing.Dict[int, typing.Dict[str, float]],
        maxima: typing.Dict[int, typing.Dict[str, float]],
    ) -> None:

        self.maxima = copy.deepcopy(maxima)
        target_indices = np.array(list(range(y.shape[1])))
        models = dict()
        weights = dict()

        self.strategies_ = copy.deepcopy(y.columns.to_list())
        y_pd = y.copy()
        y = y.to_numpy()

        for i in range(len(target_indices)):
            models[i] = dict()
            weights[i] = dict()
            for j in range(i + 1, len(target_indices)):

                if self.configuration["normalization"] in (
                    "all",
                    "binary",
                    "y",
                    "all1",
                    "binary1",
                ):
                    minimum2 = np.ones(len(X)) * np.inf
                    maximum2 = np.zeros(len(X))

                    if self.configuration["normalization"] in ("all", "all1"):
                        for idx, task_id in enumerate(X.index):
                            for method_id in range(len(target_indices)):
                                minimum2[idx] = np.nanmin(
                                    (
                                        minimum2[idx],
                                        minima[task_id][self.strategies_[method_id]],
                                    )
                                )
                                maximum2[idx] = np.nanmax(
                                    (
                                        maximum2[idx],
                                        maxima[task_id][self.strategies_[method_id]],
                                    )
                                )
                        if self.configuration["normalization"] == "all1":
                            maximum2 = np.ones_like(maximum2)
                    elif self.configuration["normalization"] in ("binary", "binary1"):
                        for idx, task_id in enumerate(X.index):
                            for method_id in (i, j):
                                minimum2[idx] = np.nanmin(
                                    (
                                        minimum2[idx],
                                        minima[task_id][self.strategies_[method_id]],
                                    )
                                )
                                maximum2[idx] = np.nanmax(
                                    (
                                        maximum2[idx],
                                        maxima[task_id][self.strategies_[method_id]],
                                    )
                                )
                        if self.configuration["normalization"] == "binary1":
                            maximum2 = np.ones_like(maximum2)
                    elif self.configuration["normalization"] == "y":
                        for idx, task_id in enumerate(X.index):
                            minimum2[idx] = np.nanmin(
                                (minimum2[idx], y_pd.loc[task_id].min())
                            )
                            maximum2[idx] = np.nanmax(
                                (maximum2[idx], y_pd.loc[task_id].max())
                            )
                    else:
                        raise ValueError(self.configuration["normalization"])

                    y_i_j = y[:, i] < y[:, j]
                    mask = np.isfinite(y[:, i]) & np.isfinite(y[:, j])
                    X_ = X.to_numpy()[mask]
                    y_i_j = y_i_j[mask]

                    minimum = minimum2[mask]
                    maximum = maximum2[mask]

                    diff = maximum - minimum
                    diff[diff == 0] = 1
                    normalized_y_i = (y[:, i][mask].copy() - minimum) / diff
                    normalized_y_j = (y[:, j][mask].copy() - minimum) / diff

                    weights_i_j = np.abs(normalized_y_i - normalized_y_j)

                elif self.configuration["normalization"] == "rank":
                    y_i_j = y[:, i] < y[:, j]
                    mask = np.isfinite(y[:, i]) & np.isfinite(y[:, j])
                    X_ = X.to_numpy()[mask]
                    y_i_j = y_i_j[mask]
                    ranks = scipy.stats.rankdata(y[mask], axis=1)
                    weights_i_j = np.abs(ranks[:, i] - ranks[:, j])

                elif self.configuration["normalization"] == "None":
                    y_i_j = y[:, i] < y[:, j]
                    mask = np.isfinite(y[:, i]) & np.isfinite(y[:, j])
                    X_ = X.to_numpy()[mask]
                    y_i_j = y_i_j[mask]
                    weights_i_j = np.ones_like(y_i_j).astype(int)

                else:
                    raise ValueError(self.configuration["normalization"])

                if len(y_i_j) == 0:
                    models[i][j] = None
                    weights[i][j] = None
                    continue

                if np.all([target == y_i_j[0] for target in y_i_j]):
                    n_zeros = int(np.ceil(len(y_i_j) / 2))
                    n_ones = int(np.floor(len(y_i_j) / 2))
                    import sklearn.dummy

                    base_model = sklearn.dummy.DummyClassifier(
                        strategy="constant", constant=y_i_j[0]
                    )
                    base_model.fit(
                        X_,
                        np.array(([[0]] * n_zeros) + ([[1]] * n_ones)).flatten(),
                        sample_weight=weights_i_j,
                    )
                else:
                    if self.configuration.get("max_depth") == 0:
                        import sklearn.dummy

                        loss_i = np.sum((y_i_j == 0) * weights_i_j)
                        loss_j = np.sum((y_i_j == 1) * weights_i_j)

                        base_model = sklearn.dummy.DummyClassifier(
                            strategy="constant",
                            constant=1 if loss_i < loss_j else 0,
                        )
                        base_model.fit(
                            X_,
                            np.ones_like(y_i_j) * int(loss_i < loss_j),
                            sample_weight=weights_i_j,
                        )
                    else:
                        base_model = self.fit_pairwise_model(
                            X_,
                            y_i_j,
                            weights_i_j,
                            self.rng,
                            self.configuration,
                        )
                models[i][j] = base_model
                weights[i][j] = weights_i_j
        self.models = models
        self.weights_ = weights
        self.target_indices = target_indices

    def _predict(
        self, X: pd.DataFrame, y: typing.Optional[pd.DataFrame]
    ) -> pd.DataFrame:

        if y is not None:
            raise ValueError("y must not be provided")

        raw_predictions = dict()
        raw_probas = dict()
        for i in range(len(self.target_indices)):
            for j in range(i + 1, len(self.target_indices)):
                if self.models[i][j] is not None:
                    raw_predictions[(i, j)] = self.models[i][j].predict(X)
                    raw_probas[(i, j)] = self.models[i][j].predict_proba(X)

        if len(raw_predictions) == 0:
            predictions = pd.DataFrame(
                0, index=X.index, columns=self.strategies_
            ).astype(float)
            predictions.iloc[:, self.single_strategy_idx] = 1.0
            return predictions

        predictions = {}
        for x_idx in range(X.shape[0]):
            wins = np.zeros(self.target_indices.shape)
            for i in range(len(self.target_indices)):
                for j in range(i + 1, len(self.target_indices)):
                    if (i, j) in raw_predictions:
                        if self.configuration["prediction"] == "soft":
                            if raw_probas[(i, j)].shape[1] == 1:
                                proba = raw_probas[(i, j)][x_idx][0]
                            else:
                                proba = raw_probas[(i, j)][x_idx][1]
                            wins[i] += proba
                            wins[j] += 1 - proba
                        elif self.configuration["prediction"] == "hard":
                            prediction = raw_predictions[(i, j)][x_idx]
                            if prediction == 1:
                                wins[i] += 1
                            else:
                                wins[j] += 1
                        else:
                            raise ValueError(self.configuration["prediction"])

            n_prev = np.inf
            # Tie breaking
            while True:
                most_wins = np.max(wins)
                most_wins_mask = most_wins == wins
                if n_prev == np.sum(most_wins_mask):
                    n_prev = np.sum(most_wins_mask)
                    hit = False
                    for method in self.tie_break_order:
                        if method not in self.strategies_:
                            continue
                        method_idx = self.strategies_.index(method)
                        if most_wins_mask[method_idx]:
                            wins[method_idx] += 1
                            hit = True
                            break
                    if not hit:
                        wins[
                            int(self.rng.choice(np.argwhere(most_wins_mask).flatten()))
                        ] += 1
                elif np.sum(most_wins_mask) > 1:
                    n_prev = np.sum(most_wins_mask)
                    where = np.argwhere(most_wins_mask).flatten()
                    for i, j in itertools.combinations(where, 2):
                        if (i, j) in raw_predictions:
                            prediction = raw_predictions[(i, j)][x_idx]
                            if prediction == 1:
                                wins[i] += 1
                            else:
                                wins[j] += 1
                        else:
                            method_i = self.strategies_[i]
                            method_j = self.strategies_[j]
                            if self.tie_break_order.index(
                                method_i
                            ) < self.tie_break_order.index(method_j):
                                wins[i] += 1
                            else:
                                wins[j] += 1
                else:
                    break

            wins = wins / np.sum(wins)
            predictions[X.index[x_idx]] = wins

        return_value = {
            task_id: {
                strategy: predictions[task_id][strategy_idx]
                for strategy_idx, strategy in enumerate(self.strategies_)
            }
            for task_id in X.index
        }
        return_value = pd.DataFrame(return_value).transpose().astype(float)
        return_value = return_value[self.strategies_]
        return_value = return_value.fillna(0.0)
        return return_value

    def fit_pairwise_model(self, X, y, weights, rng, configuration):
        raise NotImplementedError()


class OVORF(OneVSOneSelector):
    def __init__(self, n_estimators, **kwargs):
        super().__init__(**kwargs)
        self.n_estimators = n_estimators

    def fit_pairwise_model(self, X, y, weights, rng, configuration):
        base_model = sklearn.ensemble.RandomForestClassifier(
            random_state=rng,
            n_estimators=self.n_estimators,
            bootstrap=True if configuration["bootstrap"] == "True" else False,
            min_samples_split=configuration["min_samples_split"],
            min_samples_leaf=configuration["min_samples_leaf"],
            max_features=int(configuration["max_features"]),
            max_depth=configuration["max_depth"],
        )
        base_model.fit(X, y, sample_weight=weights)
        return base_model


class FallbackWrapper(AbstractSelector):
    def __init__(self, selector, default_strategies: typing.List[str]):
        self.selector = selector
        self.default_strategies = default_strategies

    def fit(
        self,
        X: pd.DataFrame,
        y: pd.DataFrame,
        minima: typing.Dict[int, typing.Dict[str, float]],
        maxima: typing.Dict[int, typing.Dict[str, float]],
    ) -> None:
        self.X_ = X
        self.strategies_ = y.columns
        self.return_value_ = np.array(
            [
                (len(self.strategies_) - self.default_strategies.index(strategy) - 1)
                / (len(self.strategies_) - 1)
                for strategy in self.strategies_
            ]
        )
        self.return_value_ = self.return_value_ / np.sum(self.return_value_)
        self.selector.fit(X, y, minima, maxima)

    def _predict(
        self, X: pd.DataFrame, y: typing.Optional[pd.DataFrame]
    ) -> pd.DataFrame:

        if y is not None:
            prediction = self.selector.predict(X, y)
        else:
            prediction = self.selector.predict(X)
        for task_id, x in X.iterrows():
            counter = 0
            te = x.copy()
            assert len(te) == self.X_.shape[1]
            for _, tr in self.X_.iterrows():
                tr = tr.to_numpy()
                if all([tr[i] >= te[i] for i in range(self.X_.shape[1])]):
                    counter += 1

            if counter == 0:
                prediction.loc[task_id] = pd.Series(
                    {
                        strategy: value
                        for strategy, value in zip(self.strategies_, self.return_value_)
                    }
                )

        return prediction



from __future__ import annotations

from typing import Any

import math
import shutil
import tempfile
import uuid
from pathlib import Path

sizes = {
    "B": 0,
    "KB": 1,
    "MB": 2,
    "GB": 3,
    "TB": 4,
}


def sizeof(path: Path | str, unit: str = "B") -> float:
    """Get the size of some path object

    Parameters
    ----------
    path : Path | str
        The path of the file or directory to get the size of

    unit : "B" | "KB" | "MB" | "GB" | "TB" = "B"
        What unit to get the answer in

    Returns
    -------
    float
        The size of the folder/file in the given units
    """
    if unit not in sizes:
        raise ValueError(f"Not a known unit {unit}")

    if not isinstance(path, Path):
        path = Path(path)

    if path.is_file():
        size = path.stat().st_size
    else:
        size = sum(f.stat().st_size for f in path.glob("**/*") if f.is_file())

    power = sizes[unit]
    return size / math.pow(1024, power)


def rmtree(
    path: Path | str,
    *,
    atomic: bool = False,
    tmp: bool | Path | str = False,
    **kwargs: Any,
) -> None:
    """Delete a file or directory

    Parameters
    ----------
    path: Path | str
        The path to delete

    atomic: bool = False
        Whether to delete the file/folder atomically. This is done by first
        using a `move` before `rmtree`.

        The `move` is not guaranteed to be atomic if moving between
        different file systems which can happen when moving to /tmp,
        depending on the OS and setup.

        The deletion part is not atomic.

        * https://docs.python.org/3/library/shutil.html#shutil.move

    tmp: bool | Path | str = False
        If bool, this defines whether atomic should use the tmp dir
        for it's move. Otherwise, a path can be specified to use

    **kwargs
        Forwarded to `rmtree`
        * https://docs.python.org/3/library/shutil.html#shutil.rmtree
    """
    if isinstance(path, str):
        path = Path(path)

    if atomic:
        if tmp is True:
            dir = Path(tempfile.gettempdir())
            uid = uuid.uuid4()
            mvpath = dir / f"autosklearn-{path.name}.old_{uid}"

        elif tmp is False:
            uid = uuid.uuid4()
            mvpath = path.parent / f"{path.name}.old_{uid}"
        else:
            mvpath = tmp if isinstance(tmp, Path) else Path(tmp)

        shutil.move(str(path), str(mvpath))
        shutil.rmtree(mvpath, **kwargs)
    else:
        shutil.rmtree(path, **kwargs)



import sklearn.metrics
from sklearn.tree import DecisionTreeClassifier

from autosklearn.pipeline.components.feature_preprocessing.polynomial import (
    PolynomialFeatures,
)
from autosklearn.pipeline.util import (
    PreprocessingTestCase,
    _test_preprocessing,
    get_dataset,
)


class PolynomialFeaturesComponentTest(PreprocessingTestCase):
    def test_default_configuration(self):
        transformation, original = _test_preprocessing(PolynomialFeatures)
        self.assertEqual(transformation.shape[0], original.shape[0])
        self.assertFalse((transformation == 0).all())

    def test_default_configuration_classify(self):
        for i in range(2):
            X_train, Y_train, X_test, Y_test = get_dataset(
                dataset="breast_cancer", make_sparse=False
            )
            configuration_space = PolynomialFeatures.get_hyperparameter_search_space()
            default = configuration_space.get_default_configuration()
            preprocessor = PolynomialFeatures(
                random_state=1, **{hp_name: default[hp_name] for hp_name in default}
            )
            preprocessor.fit(X_train, Y_train)
            X_train_trans = preprocessor.transform(X_train)
            X_test_trans = preprocessor.transform(X_test)

            # fit a classifier on top
            classifier = DecisionTreeClassifier(random_state=1)
            predictor = classifier.fit(X_train_trans, Y_train)
            predictions = predictor.predict(X_test_trans)
            accuracy = sklearn.metrics.accuracy_score(predictions, Y_test)
            self.assertAlmostEqual(accuracy, 0.9474940334128878, places=2)

    def test_default_configuration_classify_sparse(self):
        for i in range(2):
            X_train, Y_train, X_test, Y_test = get_dataset(
                dataset="breast_cancer", make_sparse=True
            )
            configuration_space = PolynomialFeatures.get_hyperparameter_search_space()
            default = configuration_space.get_default_configuration()
            preprocessor = PolynomialFeatures(
                random_state=1, **{hp_name: default[hp_name] for hp_name in default}
            )
            preprocessor.fit(X_train, Y_train)
            X_train_trans = preprocessor.transform(X_train)
            X_test_trans = preprocessor.transform(X_test)

            # fit a classifier on top
            classifier = DecisionTreeClassifier(random_state=1)
            predictor = classifier.fit(X_train_trans, Y_train)
            predictions = predictor.predict(X_test_trans)
            accuracy = sklearn.metrics.accuracy_score(predictions, Y_test)
            self.assertAlmostEqual(accuracy, 0.8544152744630071, places=2)

    def test_preprocessing_dtype(self):
        super(PolynomialFeaturesComponentTest, self)._test_preprocessing_dtype(
            PolynomialFeatures, test_sparse=False
        )



from typing import Any, Dict, Optional, Union

import copy
import itertools
import os
import resource
import tempfile

import numpy as np
import sklearn.datasets
import sklearn.decomposition
import sklearn.ensemble
import sklearn.model_selection
import sklearn.svm
from ConfigSpace.configuration_space import ConfigurationSpace
from ConfigSpace.hyperparameters import CategoricalHyperparameter
from joblib import Memory
from sklearn.base import clone
from sklearn.utils.validation import check_is_fitted

import autosklearn.pipeline.components.classification as classification_components
import autosklearn.pipeline.components.feature_preprocessing as preprocessing_components
from autosklearn.askl_typing import FEAT_TYPE_TYPE
from autosklearn.pipeline.classification import SimpleClassificationPipeline
from autosklearn.pipeline.components.base import (
    AutoSklearnChoice,
    AutoSklearnClassificationAlgorithm,
    AutoSklearnComponent,
    AutoSklearnPreprocessingAlgorithm,
    _addons,
)
from autosklearn.pipeline.constants import (
    DENSE,
    INPUT,
    PREDICTIONS,
    SIGNED_DATA,
    SPARSE,
    UNSIGNED_DATA,
)
from autosklearn.pipeline.util import get_dataset

import unittest
import unittest.mock

from test.test_pipeline.ignored_warnings import classifier_warnings, ignore_warnings


class DummyClassifier(AutoSklearnClassificationAlgorithm):
    @staticmethod
    def get_properties(dataset_properties=None):
        return {
            "shortname": "AB",
            "name": "AdaBoost Classifier",
            "handles_regression": False,
            "handles_classification": True,
            "handles_multiclass": True,
            "handles_multilabel": True,
            "handles_multioutput": False,
            "is_deterministic": True,
            "input": (DENSE, SPARSE, UNSIGNED_DATA),
            "output": (PREDICTIONS,),
        }

    @staticmethod
    def get_hyperparameter_search_space(feat_type=None, dataset_properties=None):
        cs = ConfigurationSpace()
        return cs


class DummyPreprocessor(AutoSklearnPreprocessingAlgorithm):
    @staticmethod
    def get_properties(dataset_properties=None):
        return {
            "shortname": "AB",
            "name": "AdaBoost Classifier",
            "handles_regression": False,
            "handles_classification": True,
            "handles_multiclass": True,
            "handles_multilabel": True,
            "handles_multioutput": False,
            "is_deterministic": True,
            "input": (DENSE, SPARSE, UNSIGNED_DATA),
            "output": (INPUT,),
        }

    @staticmethod
    def get_hyperparameter_search_space(dataset_properties=None):
        cs = ConfigurationSpace()
        return cs


class CrashPreprocessor(AutoSklearnPreprocessingAlgorithm):
    def __init__(*args, **kwargs):
        pass

    @staticmethod
    def get_properties(dataset_properties=None):
        return {
            "shortname": "AB",
            "name": "AdaBoost Classifier",
            "handles_regression": False,
            "handles_classification": True,
            "handles_multiclass": True,
            "handles_multilabel": True,
            "handles_multioutput": False,
            "is_deterministic": True,
            "input": (DENSE, SPARSE, UNSIGNED_DATA),
            "output": (INPUT,),
        }

    def fit(self, X, y):
        raise ValueError("Make sure fit is called")

    @staticmethod
    def get_hyperparameter_search_space(dataset_properties=None):
        cs = ConfigurationSpace()
        return cs


class SimpleClassificationPipelineTest(unittest.TestCase):
    _multiprocess_can_split_ = True

    def test_io_dict(self):
        """Test for the properties of classifier components

        Expects
        -------
        * All required properties are stated in class `get_properties()`
        """
        classifiers = classification_components._classifiers
        for c in classifiers:
            if classifiers[c] == classification_components.ClassifierChoice:
                continue
            props = classifiers[c].get_properties()
            self.assertIn("input", props)
            self.assertIn("output", props)
            inp = props["input"]
            output = props["output"]

            self.assertIsInstance(inp, tuple)
            self.assertIsInstance(output, tuple)
            for i in inp:
                self.assertIn(i, (SPARSE, DENSE, SIGNED_DATA, UNSIGNED_DATA))
            self.assertEqual(output, (PREDICTIONS,))
            self.assertIn("handles_regression", props)
            self.assertFalse(props["handles_regression"])
            self.assertIn("handles_classification", props)
            self.assertIn("handles_multiclass", props)
            self.assertIn("handles_multilabel", props)

    def test_find_classifiers(self):
        """Test that the classifier components can be found

        Expects
        -------
        * At least two classifier components can be found
        * They inherit from AutoSklearnClassificationAlgorithm
        """
        classifiers = classification_components._classifiers
        self.assertGreaterEqual(len(classifiers), 2)
        for key in classifiers:
            if hasattr(classifiers[key], "get_components"):
                continue
            self.assertIn(
                AutoSklearnClassificationAlgorithm, classifiers[key].__bases__
            )

    def test_find_preprocessors(self):
        """Test that preproccesor components can be found

        Expects
        -------
        * At least 1 preprocessor component can be found
        * The inherit from AutoSklearnPreprocessingAlgorithm
        """
        preprocessors = preprocessing_components._preprocessors
        self.assertGreaterEqual(len(preprocessors), 1)
        for key in preprocessors:
            if hasattr(preprocessors[key], "get_components"):
                continue
            self.assertIn(
                AutoSklearnPreprocessingAlgorithm, preprocessors[key].__bases__
            )

    def test_default_configuration(self):
        """Test that seeded SimpleClassificaitonPipeline returns good results on iris

        Expects
        -------
        * The performance of configuration with fixed seed gets above 96% accuracy
            on iris
        """
        X_train, Y_train, X_test, Y_test = get_dataset(dataset="iris")

        auto = SimpleClassificationPipeline(random_state=1)

        with ignore_warnings(classifier_warnings):
            auto = auto.fit(X_train, Y_train)

        predictions = auto.predict(X_test)

        acc = sklearn.metrics.accuracy_score(predictions, Y_test)
        self.assertAlmostEqual(0.96, acc)

    def test_default_configuration_multilabel(self):
        """Test that SimpleClassificationPipeline default config returns good results on
        a multilabel version of iris.

        Expects
        -------
        * The performance of a random configuratino gets above 96% on a multilabel
            version of iris
        """
        X_train, Y_train, X_test, Y_test = get_dataset(
            dataset="iris", make_multilabel=True
        )

        classifier = SimpleClassificationPipeline(
            dataset_properties={"multilabel": True}, random_state=0
        )
        cs = classifier.get_hyperparameter_search_space()

        default = cs.get_default_configuration()
        classifier.set_hyperparameters(default)

        with ignore_warnings(classifier_warnings):
            classifier = classifier.fit(X_train, Y_train)

        predictions = classifier.predict(X_test)

        acc = sklearn.metrics.accuracy_score(predictions, Y_test)
        self.assertAlmostEqual(0.96, acc)

    def test_default_configuration_iterative_fit(self):
        """Test that the SimpleClassificationPipeline default config for random forest
        with no preprocessing can be iteratively fit on iris.

        Expects
        -------
        * Random forest pipeline can be fit iteratively
        * Test that its number of estimators is equal to the iteration count
        """
        X_train, Y_train, X_test, Y_test = get_dataset(dataset="iris")

        classifier = SimpleClassificationPipeline(
            include={
                "classifier": ["random_forest"],
                "feature_preprocessor": ["no_preprocessing"],
            },
            random_state=0,
        )
        classifier.fit_transformer(X_train, Y_train)

        with ignore_warnings(classifier_warnings):
            for i in range(1, 11):
                classifier.iterative_fit(X_train, Y_train)
                n_estimators = classifier.steps[-1][-1].choice.estimator.n_estimators
                self.assertEqual(n_estimators, i)

    def test_repr(self):
        """Test that the default pipeline can be converted to its representation and
        converted back.

        Expects
        -------
        * The the SimpleClassificationPipeline has a repr
        * This repr can be evaluated back to an instance of SimpleClassificationPipeline
        """
        representation = repr(SimpleClassificationPipeline())
        cls = eval(representation)
        self.assertIsInstance(cls, SimpleClassificationPipeline)

    def test_multilabel(self):
        """Test non-seeded configurations for multi-label data

        Expects
        -------
        * All configurations should fit, predict and predict_proba successfully
        """
        cache = Memory(location=tempfile.gettempdir())
        cached_func = cache.cache(sklearn.datasets.make_multilabel_classification)
        X, Y = cached_func(
            n_samples=150,
            n_features=20,
            n_classes=5,
            n_labels=2,
            length=50,
            allow_unlabeled=True,
            sparse=False,
            return_indicator=True,
            return_distributions=False,
            random_state=1,
        )

        data = {
            "X_train": X[:100, :],
            "Y_train": Y[:100, :],
            "X_test": X[101:, :],
            "Y_test": Y[
                101:,
            ],
        }

        pipeline = SimpleClassificationPipeline(dataset_properties={"multilabel": True})
        cs = pipeline.get_hyperparameter_search_space()
        self._test_configurations(configurations_space=cs, dataset=data)

    def test_configurations(self):
        """Test non-seeded random sets of configurations with default dataset properties

        Expects
        -------
        * All configurations should fit, predict and predict_proba successfully
        """
        cls = SimpleClassificationPipeline()
        cs = cls.get_hyperparameter_search_space()
        self._test_configurations(configurations_space=cs)

    def test_configurations_signed_data(self):
        """Tests a non-seeded random set of configurations with signed data

        Expects
        -------
        * All configurations should fit, predict and predict_proba successfully
        """
        dataset_properties = {"signed": True}

        cls = SimpleClassificationPipeline(dataset_properties=dataset_properties)
        cs = cls.get_hyperparameter_search_space()

        self._test_configurations(
            configurations_space=cs, dataset_properties=dataset_properties
        )

    def test_configurations_sparse(self):
        """Tests a non-seeded random set of configurations with sparse data

        Expects
        -------
        * All configurations should fit, predict and predict_proba successfully
        """
        pipeline = SimpleClassificationPipeline(dataset_properties={"sparse": True})
        cs = pipeline.get_hyperparameter_search_space()

        self._test_configurations(configurations_space=cs, make_sparse=True)

    def test_configurations_categorical_data(self):
        """Tests a non-seeded random set of configurations with sparse, mixed data

        Loads specific data from <here>/components/data_preprocessing/dataset.pkl

        Expects
        -------
        * All configurations should fit, predict and predict_proba successfully
        """
        categorical_columns = [
            True,
            True,
            True,
            False,
            False,
            True,
            True,
            True,
            False,
            True,
            True,
            True,
            True,
            True,
            True,
            True,
            True,
            True,
            True,
            True,
            True,
            True,
            True,
            True,
            True,
            True,
            True,
            True,
            True,
            True,
            True,
            True,
            False,
            False,
            False,
            True,
            True,
            True,
        ]
        categorical = {
            i: "categorical" if is_categorical else "numerical"
            for i, is_categorical in enumerate(categorical_columns)
        }

        pipeline = SimpleClassificationPipeline(
            feat_type=categorical,
            dataset_properties={"sparse": False},
            include={
                "feature_preprocessor": ["no_preprocessing"],
                "classifier": ["sgd", "adaboost"],
            },
        )

        cs = pipeline.get_hyperparameter_search_space(feat_type=categorical)

        here = os.path.dirname(__file__)
        dataset_path = os.path.join(
            here, "components", "data_preprocessing", "dataset.pkl"
        )

        X = np.loadtxt(dataset_path)
        y = X[:, -1].copy()
        X = X[:, :-1]
        X_train, X_test, Y_train, Y_test = sklearn.model_selection.train_test_split(
            X, y
        )

        data = {
            "X_train": X_train,
            "Y_train": Y_train,
            "X_test": X_test,
            "Y_test": Y_test,
        }

        init_params = {"data_preprocessor:feat_type": categorical}

        self._test_configurations(
            configurations_space=cs,
            dataset=data,
            init_params=init_params,
            feat_type=categorical,
        )

    @unittest.mock.patch(
        "autosklearn.pipeline.components.data_preprocessing"
        ".DataPreprocessorChoice.set_hyperparameters"
    )
    def test_categorical_passed_to_one_hot_encoder(self, ohe_mock):
        """Test that the feat_types arg is passed to the OneHotEncoder

        Expects
        -------
        * Construction of SimpleClassificationPipeline to pass init_params correctly
            to the OneHotEncoder

        * Setting the pipeline's hyperparameters after construction also correctly
            sets the init params of the OneHotEncoder
        """

        # Mock the _check_init_params_honored as there is no object created,
        # _check_init_params_honored will fail as a datapreprocessor was never created
        with unittest.mock.patch(
            "autosklearn.pipeline.classification.SimpleClassificationPipeline"
            "._check_init_params_honored"
        ):

            # Check through construction
            feat_types = {0: "categorical", 1: "numerical"}

            cls = SimpleClassificationPipeline(
                feat_type=feat_types,
                init_params={"data_preprocessor:feat_type": feat_types},
            )

            init_args = ohe_mock.call_args[1]["init_params"]
            self.assertEqual(init_args, {"feat_type": feat_types})

            # Check through `set_hyperparameters`
            feat_types = {0: "categorical", 1: "categorical", 2: "numerical"}

            default = cls.get_hyperparameter_search_space(
                feat_type=feat_types
            ).get_default_configuration()
            cls.set_hyperparameters(
                feat_type=feat_types,
                configuration=default,
                init_params={"data_preprocessor:feat_type": feat_types},
            )

            init_args = ohe_mock.call_args[1]["init_params"]
            self.assertEqual(init_args, {"feat_type": feat_types})

    def _test_configurations(
        self,
        configurations_space: ConfigurationSpace,
        make_sparse: bool = False,
        dataset: Union[str, Dict[str, Any]] = "digits",
        init_params: Dict[str, Any] = None,
        dataset_properties: Dict[str, Any] = None,
        n_samples: int = 10,
        feat_type: Optional[FEAT_TYPE_TYPE] = None,
    ):
        """Tests a configuration space by taking multiple samples and fiting each
        before calling predict and predict_proba.

        Parameters
        ----------
        configurations_space: ConfigurationSpace
            The configuration space to sample from

        make_sparse: bool = False
            Whether to make the dataset sparse or not

        dataset: Union[str, Dict[str, Any]] = 'digits'
            Either a dataset name or a dictionary as below. If given a str, it will
            use `make_sparse` and add NaNs to the dataset.

            {'X_train': ..., 'Y_train': ..., 'X_test': ..., 'y_test': ...}

        init_params: Dict[str, Any] = None
            A dictionary of initial parameters to give to the pipeline.

        dataset_properties: Dict[str, Any]
            A dictionary of properties describing the dataset

        n_samples: int = 10
            How many configurations to sample
        """
        # Use a limit of ~3GiB
        limit = 3072 * 1024 * 1024
        resource.setrlimit(resource.RLIMIT_AS, (limit, limit))

        for i in range(n_samples):
            config = configurations_space.sample_configuration()
            config._populate_values()

            # Restrict configurations which could take too long on travis-ci
            restrictions = {
                "classifier:passive_aggressive:n_iter": 5,
                "classifier:sgd:n_iter": 5,
                "classifier:adaboost:n_estimators": 50,
                "classifier:adaboost:max_depth": 1,
                "feature_preprocessor:kernel_pca:n_components": 10,
                "feature_preprocessor:kitchen_sinks:n_components": 50,
                "classifier:proj_logit:max_epochs": 1,
                "classifier:libsvm_svc:degree": 2,
                "regressor:libsvm_svr:degree": 2,
                "feature_preprocessor:truncatedSVD:target_dim": 10,
                "feature_preprocessor:polynomial:degree": 2,
                "classifier:lda:n_components": 10,
                "feature_preprocessor:nystroem_sampler:n_components": 50,
                "feature_preprocessor:feature_agglomeration:n_clusters": 2,
                "classifier:gradient_boosting:max_leaf_nodes": 64,
            }

            config._values.update(
                {
                    param: value
                    for param, value in restrictions.items()
                    if param in config and config[param] is not None
                }
            )

            if isinstance(dataset, str):
                X_train, Y_train, X_test, Y_test = get_dataset(
                    dataset=dataset, make_sparse=make_sparse, add_NaNs=True
                )
            else:
                X_train = dataset["X_train"].copy()
                Y_train = dataset["Y_train"].copy()
                X_test = dataset["X_test"].copy()
                dataset["Y_test"].copy()

            init_params_ = copy.deepcopy(init_params)

            cls = SimpleClassificationPipeline(
                feat_type=feat_type,
                dataset_properties=dataset_properties,
                init_params=init_params_,
            )
            cls.set_hyperparameters(
                config, init_params=init_params_, feat_type=feat_type
            )

            # First make sure that for this configuration, setting the parameters
            # does not mistakenly set the estimator as fitted
            for name, step in cls.named_steps.items():
                with self.assertRaisesRegex(
                    sklearn.exceptions.NotFittedError, "instance is not fitted yet"
                ):
                    check_is_fitted(step)

            try:
                with ignore_warnings(classifier_warnings):
                    cls.fit(X_train, Y_train)

                # After fit, all components should be tagged as fitted
                # by sklearn. Check is fitted raises an exception if that
                # is not the case
                try:
                    for name, step in cls.named_steps.items():
                        check_is_fitted(step)
                except sklearn.exceptions.NotFittedError:
                    self.fail(f"config={config} raised NotFittedError unexpectedly!")

                cls.predict(X_test.copy())
                cls.predict_proba(X_test)

            except MemoryError:
                continue
            except np.linalg.LinAlgError:
                continue
            except ValueError as e:
                if "Floating-point under-/overflow occurred at epoch" in e.args[0]:
                    continue
                elif "removed all features" in e.args[0]:
                    continue
                elif "all features are discarded" in e.args[0]:
                    continue
                elif "Numerical problems in QDA" in e.args[0]:
                    continue
                elif "Bug in scikit-learn" in e.args[0]:
                    continue
                elif (
                    "The condensed distance matrix must contain only finite "
                    "values." in e.args[0]
                ):
                    continue
                elif "Internal work array size computation failed" in e.args[0]:
                    continue
                # Assumed to be caused by knn with preprocessor fast_ica with whiten
                elif "Input contains NaN, infinity or a value too large" in e.args[0]:
                    continue
                elif (
                    "zero-size array to reduction operation maximum which has no "
                    "identity" in e.args[0]
                ):
                    continue
                else:
                    e.args += (f"config={config}",)
                    raise e

            except RuntimeWarning as e:
                if "invalid value encountered in sqrt" in e.args[0]:
                    continue
                elif "invalid value encountered in multiply" in e.args[0]:
                    continue
                elif "divide by zero encountered in" in e.args[0]:
                    continue
                elif "invalid value encountered in divide" in e.args[0]:
                    continue
                elif "invalid value encountered in true_divide" in e.args[0]:
                    continue
                elif "invalid value encountered in multiply" in e.args[0]:
                    continue
                else:
                    e.args += (f"config={config}",)
                    raise e

            except UserWarning as e:
                if "FastICA did not converge" in e.args[0]:
                    continue
                else:
                    e.args += (f"config={config}",)
                    raise e

    def test_get_hyperparameter_search_space(self):
        """Test the configuration space returned by a SimpleClassificationPipeline

        Expects
        -------
        * pipeline returns a configurations space
        * 7 rescaling choices
        * 16 classifier choices
        * 13 features preprocessor choices
        * 176 total hyperparameters
        * (n_hyperparameters - 4) different conditionals for the pipeline
        * 53 forbidden combinations
        """
        pipeline = SimpleClassificationPipeline(
            feat_type={"A": "numerical", "B": "categorical", "C": "string"}
        )
        cs = pipeline.get_hyperparameter_search_space()
        self.assertIsInstance(cs, ConfigurationSpace)

        rescale_param = (
            "data_preprocessor:feature_type:numerical_transformer:rescaling:__choice__"
        )
        n_choices = len(cs.get_hyperparameter(rescale_param).choices)
        self.assertEqual(n_choices, 7)

        n_classifiers = len(cs.get_hyperparameter("classifier:__choice__").choices)
        self.assertEqual(n_classifiers, 16)

        n_preprocessors = len(
            cs.get_hyperparameter("feature_preprocessor:__choice__").choices
        )
        self.assertEqual(n_preprocessors, 13)

        hyperparameters = cs.get_hyperparameters()
        self.assertEqual(len(hyperparameters), 179)

        # for hp in sorted([str(h) for h in hyperparameters]):
        #    print hp

        # The four components which are always active are classifier,
        # feature preprocessor, balancing and data preprocessing pipeline.
        conditions = cs.get_conditions()
        self.assertEqual(len(hyperparameters) - 4, len(conditions))

        forbiddens = cs.get_forbiddens()
        self.assertEqual(len(forbiddens), 53)

    def test_get_hyperparameter_search_space_include_exclude_models(self):
        """Test the configuration space when using include and exclude

        Expects
        -------
        * Including a classifier has pipeline give back matching choice
        * Excluding a classifier means it won't show up in the hyperparameter space
        * Including a feature preprocessor has pipeline give back matching choice
        * Excluding a feature preprocessor means it won't show up in the
            hyperparameter space
        """
        # include a classifier choice
        pipeline = SimpleClassificationPipeline(include={"classifier": ["libsvm_svc"]})
        cs = pipeline.get_hyperparameter_search_space()

        expected = CategoricalHyperparameter("classifier:__choice__", ["libsvm_svc"])
        returned = cs.get_hyperparameter("classifier:__choice__")
        self.assertEqual(returned, expected)

        # exclude a classifier choice
        pipeline = SimpleClassificationPipeline(exclude={"classifier": ["libsvm_svc"]})
        cs = pipeline.get_hyperparameter_search_space()
        self.assertNotIn("libsvm_svc", str(cs))

        # include a feature preprocessor
        pipeline = SimpleClassificationPipeline(
            include={"feature_preprocessor": ["select_percentile_classification"]}
        )
        cs = pipeline.get_hyperparameter_search_space()

        returned = cs.get_hyperparameter("feature_preprocessor:__choice__")
        expected = CategoricalHyperparameter(
            "feature_preprocessor:__choice__", ["select_percentile_classification"]
        )
        self.assertEqual(returned, expected)

        # exclude a feature preprocessor
        pipeline = SimpleClassificationPipeline(
            exclude={"feature_preprocessor": ["select_percentile_classification"]}
        )
        cs = pipeline.get_hyperparameter_search_space()
        self.assertNotIn("select_percentile_classification", str(cs))

    def test_get_hyperparameter_search_space_preprocessor_contradicts_default(
        self,
    ):
        """Test that the default classifier gets updated based on the legal feature
        preprocessors that come before.

        Expects
        -------
        * With 'densifier' as only legal feature_preprocessor, 'qda' is default
        * With 'nystroem_sampler' as only legal feature_preprocessor, 'sgd' is default
        """
        pipeline = SimpleClassificationPipeline(
            include={"feature_preprocessor": ["densifier"]},
            dataset_properties={"sparse": True},
        )
        cs = pipeline.get_hyperparameter_search_space()

        default_choice = cs.get_hyperparameter("classifier:__choice__").default_value
        self.assertEqual(default_choice, "qda")

        pipeline = SimpleClassificationPipeline(
            include={"feature_preprocessor": ["nystroem_sampler"]}
        )
        cs = pipeline.get_hyperparameter_search_space()

        default_choice = cs.get_hyperparameter("classifier:__choice__").default_value
        self.assertEqual(default_choice, "sgd")

    def test_get_hyperparameter_search_space_only_forbidden_combinations(self):
        """Test that invalid pipeline configurations raise errors

        Expects
        -------
        * 0 combinations are found with 'multinomial_nb' and 'pca' with 'sparse' data
        * Classifiers that can handle sparse but located behind a 'densifier' should
            raise that no legal default configuration can be found
        """
        with self.assertRaisesRegex(AssertionError, "No valid pipeline found."):
            SimpleClassificationPipeline(
                include={
                    "classifier": ["multinomial_nb"],
                    "feature_preprocessor": ["pca"],
                },
                dataset_properties={"sparse": True},
            )

        with self.assertRaisesRegex(
            ValueError, "Cannot find a legal default configuration."
        ):
            SimpleClassificationPipeline(
                include={
                    "classifier": ["liblinear_svc"],
                    "feature_preprocessor": ["densifier"],
                },
                dataset_properties={"sparse": True},
            )

    @unittest.skip("Wait until ConfigSpace is fixed.")
    def test_get_hyperparameter_search_space_dataset_properties(self):
        cs_mc = SimpleClassificationPipeline.get_hyperparameter_search_space(
            dataset_properties={"multiclass": True}
        )
        self.assertNotIn("bernoulli_nb", str(cs_mc))

        cs_ml = SimpleClassificationPipeline.get_hyperparameter_search_space(
            dataset_properties={"multilabel": True}
        )
        self.assertNotIn("k_nearest_neighbors", str(cs_ml))
        self.assertNotIn("liblinear", str(cs_ml))
        self.assertNotIn("libsvm_svc", str(cs_ml))
        self.assertNotIn("sgd", str(cs_ml))

        cs_sp = SimpleClassificationPipeline.get_hyperparameter_search_space(
            dataset_properties={"sparse": True}
        )
        self.assertIn("extra_trees", str(cs_sp))
        self.assertIn("gradient_boosting", str(cs_sp))
        self.assertIn("random_forest", str(cs_sp))

        cs_mc_ml = SimpleClassificationPipeline.get_hyperparameter_search_space(
            dataset_properties={"multilabel": True, "multiclass": True}
        )
        self.assertEqual(cs_ml, cs_mc_ml)

    def test_predict_batched(self):
        """Test that predict_proba predicts the same as the underlying classifier with
        predict_proba argument `batches`.

        Expects
        -------
        * Should expect the output shape to match that of the digits dataset
        * Should expect a fixed call count each test run
        * Should expect predict_proba with `batches` and predict_proba
            perform near identically
        """
        cls = SimpleClassificationPipeline(include={"classifier": ["sgd"]})

        # Multiclass
        X_train, Y_train, X_test, Y_test = get_dataset(dataset="digits")

        with ignore_warnings(classifier_warnings):
            cls.fit(X_train, Y_train)

        X_test_ = X_test.copy()
        prediction_ = cls.predict_proba(X_test_)

        # The object behind the last step in the pipeline
        cls_predict = unittest.mock.Mock(wraps=cls.steps[-1][1].predict_proba)
        cls.steps[-1][-1].predict_proba = cls_predict

        prediction = cls.predict_proba(X_test, batch_size=20)

        self.assertEqual((1647, 10), prediction.shape)
        self.assertEqual(84, cls_predict.call_count)
        np.testing.assert_array_almost_equal(prediction_, prediction)

    def test_predict_batched_sparse(self):
        """Test that predict_proba predicts the same as the underlying classifier with
        predict_proba argument `batches`, with a sparse dataset

        Expects
        -------
        * Should expect the output shape to match that of the digits dataset
        * Should expect a fixed call count each test run
        * Should expect predict_proba with `batches` and predict_proba
            perform near identically
        """
        cls = SimpleClassificationPipeline(
            dataset_properties={"sparse": True}, include={"classifier": ["sgd"]}
        )

        # Multiclass
        X_train, Y_train, X_test, Y_test = get_dataset(
            dataset="digits", make_sparse=True
        )
        with ignore_warnings(classifier_warnings):
            cls.fit(X_train, Y_train)

        X_test_ = X_test.copy()
        prediction_ = cls.predict_proba(X_test_)

        # The object behind the last step in the pipeline
        cls_predict = unittest.mock.Mock(wraps=cls.steps[-1][1].predict_proba)
        cls.steps[-1][-1].predict_proba = cls_predict

        prediction = cls.predict_proba(X_test, batch_size=20)

        self.assertEqual((1647, 10), prediction.shape)
        self.assertEqual(84, cls_predict.call_count)
        np.testing.assert_array_almost_equal(prediction_, prediction)

    def test_predict_proba_batched(self):
        """Test that predict_proba predicts the same as the underlying classifier with
        predict_proba argument `batches`, for multiclass and multilabel data.

        Expects
        -------
        * Should expect the output shape to match that of the digits dataset
        * Should expect a fixed call count each test run
        * Should expect predict_proba with `batches` and predict_proba
            perform near identically
        """
        # Multiclass
        cls = SimpleClassificationPipeline(
            feat_type={i: "numerical" for i in range(0, 64)},
            include={"classifier": ["sgd"]},
        )
        X_train, Y_train, X_test, Y_test = get_dataset(dataset="digits")

        with ignore_warnings(classifier_warnings):
            cls.fit(X_train, Y_train)

        X_test_ = X_test.copy()
        prediction_ = cls.predict_proba(X_test_)

        # The object behind the last step in the pipeline
        cls_predict = unittest.mock.Mock(wraps=cls.steps[-1][1].predict_proba)
        cls.steps[-1][-1].predict_proba = cls_predict

        prediction = cls.predict_proba(X_test, batch_size=20)

        self.assertEqual((1647, 10), prediction.shape)
        self.assertEqual(84, cls_predict.call_count)
        np.testing.assert_array_almost_equal(prediction_, prediction)

        # Multilabel
        cls = SimpleClassificationPipeline(include={"classifier": ["lda"]})
        X_train, Y_train, X_test, Y_test = get_dataset(dataset="digits")
        Y_train = np.array(
            list([(list([1 if i != y else 0 for i in range(10)])) for y in Y_train])
        )

        with ignore_warnings(classifier_warnings):
            cls.fit(X_train, Y_train)

        X_test_ = X_test.copy()
        prediction_ = cls.predict_proba(X_test_)

        # The object behind the last step in the pipeline
        cls_predict = unittest.mock.Mock(wraps=cls.steps[-1][1].predict_proba)
        cls.steps[-1][-1].predict_proba = cls_predict

        prediction = cls.predict_proba(X_test, batch_size=20)

        self.assertEqual((1647, 10), prediction.shape)
        self.assertEqual(84, cls_predict.call_count)
        np.testing.assert_array_almost_equal(prediction_, prediction)

    def test_predict_proba_batched_sparse(self):
        """Test that predict_proba predicts the same as the underlying classifier with
        predict_proba argument `batches`, for multiclass and multilabel data.

        Expects
        -------
        * Should expect the output shape to match that of the digits dataset
        * Should expect a fixed call count each test run
        * Should expect predict_proba with `batches` and predict_proba
            perform near identically
        """
        cls = SimpleClassificationPipeline(
            feat_type={i: "numerical" for i in range(0, 64)},
            dataset_properties={"sparse": True, "multiclass": True},
            include={"classifier": ["sgd"]},
        )

        # Multiclass
        X_train, Y_train, X_test, Y_test = get_dataset(
            dataset="digits", make_sparse=True
        )
        X_test_ = X_test.copy()

        with ignore_warnings(classifier_warnings):
            cls.fit(X_train, Y_train)

        prediction_ = cls.predict_proba(X_test_)

        # The object behind the last step in the pipeline
        cls_predict = unittest.mock.Mock(wraps=cls.steps[-1][1].predict_proba)
        cls.steps[-1][-1].predict_proba = cls_predict

        prediction = cls.predict_proba(X_test, batch_size=20)

        self.assertEqual((1647, 10), prediction.shape)
        self.assertEqual(84, cls_predict.call_count)
        np.testing.assert_array_almost_equal(prediction_, prediction)

        # Multilabel
        cls = SimpleClassificationPipeline(
            dataset_properties={"sparse": True, "multilabel": True},
            include={"classifier": ["lda"]},
        )
        X_train, Y_train, X_test, Y_test = get_dataset(
            dataset="digits", make_sparse=True
        )

        X_test_ = X_test.copy()
        Y_train = np.array([[1 if i != y else 0 for i in range(10)] for y in Y_train])

        with ignore_warnings(classifier_warnings):
            cls.fit(X_train, Y_train)

        prediction_ = cls.predict_proba(X_test_)

        # The object behind the last step in the pipeline
        cls_predict = unittest.mock.Mock(wraps=cls.steps[-1][1].predict_proba)
        cls.steps[-1][-1].predict_proba = cls_predict

        prediction = cls.predict_proba(X_test, batch_size=20)

        self.assertEqual((1647, 10), prediction.shape)
        self.assertEqual(84, cls_predict.call_count)
        np.testing.assert_array_almost_equal(prediction_, prediction)

    def test_pipeline_clonability(self):
        """Test that the pipeline item is clonable with `sklearn.clone`

        Expects
        -------
        * The cloned object has all the same param keys
        * The cloned object can be constructed from theses params
        * The reconstructed clone and the original have the same param values
        """
        X_train, Y_train, X_test, Y_test = get_dataset(dataset="iris")

        auto = SimpleClassificationPipeline()

        with ignore_warnings(classifier_warnings):
            auto = auto.fit(X_train, Y_train)

        auto_clone = clone(auto)
        auto_clone_params = auto_clone.get_params()

        # Make sure all keys are copied properly
        for k, v in auto.get_params().items():
            self.assertIn(k, auto_clone_params)

        # Make sure the params getter of estimator are honored
        klass = auto.__class__
        new_object_params = auto.get_params(deep=False)
        for name, param in new_object_params.items():
            new_object_params[name] = clone(param, safe=False)

        new_object = klass(**new_object_params)
        params_set = new_object.get_params(deep=False)

        for name in new_object_params:
            param1 = new_object_params[name]
            param2 = params_set[name]
            self.assertEqual(param1, param2)

    def test_set_params(self):
        pass

    def test_get_params(self):
        pass

    def test_add_classifier(self):
        """Test that classifiers can be added

        Expects
        -------
        * There should be 0 components initially
        * There should be 1 component after adding a classifier
        * The classifier should be in the search space of the Pipeline after being added
        """
        self.assertEqual(
            len(classification_components.additional_components.components), 0
        )
        self.assertEqual(len(_addons["classification"].components), 0)

        classification_components.add_classifier(DummyClassifier)

        self.assertEqual(
            len(classification_components.additional_components.components), 1
        )
        self.assertEqual(len(_addons["classification"].components), 1)

        cs = SimpleClassificationPipeline().get_hyperparameter_search_space()
        self.assertIn("DummyClassifier", str(cs))

        del classification_components.additional_components.components[
            "DummyClassifier"
        ]

    def test_add_preprocessor(self):
        """Test that preprocessors can be added

        Expects
        -------
        * There should be 0 components initially
        * There should be 1 component after adding a preprocessor
        * The preprocessor ii in the search space of the Pipeline after being added
        """
        self.assertEqual(
            len(preprocessing_components.additional_components.components), 0
        )
        self.assertEqual(len(_addons["feature_preprocessing"].components), 0)

        preprocessing_components.add_preprocessor(DummyPreprocessor)

        self.assertEqual(
            len(preprocessing_components.additional_components.components), 1
        )
        self.assertEqual(len(_addons["feature_preprocessing"].components), 1)

        cs = SimpleClassificationPipeline().get_hyperparameter_search_space()
        self.assertIn("DummyPreprocessor", str(cs))

        del preprocessing_components.additional_components.components[
            "DummyPreprocessor"
        ]

    def _test_set_hyperparameter_choice(
        self, expected_key, implementation, config_dict
    ):
        """Given a configuration in config, this procedure makes sure that the given
        implementation, which should be a Choice component, honors the type of the
        object, and any hyperparameter associated to it

        TODO: typing
        """
        keys_checked = [expected_key]
        implementation_type = config_dict[expected_key]
        expected_type = implementation.get_components()[implementation_type]
        self.assertIsInstance(implementation.choice, expected_type)

        # Are there further hyperparams?
        # A choice component might have attribute requirements that we need to check
        expected_sub_key = (
            expected_key.replace(":__choice__", ":") + implementation_type
        )
        expected_attributes = {}
        if "data_preprocessor:__choice__" in expected_key:
            # We have to check both the numerical and categorical
            to_check = {
                "numerical_transformer": implementation.choice.numer_ppl.named_steps,
                "categorical_transformer": implementation.choice.categ_ppl.named_steps,
                "text_transformer": implementation.choice.txt_ppl.named_steps,
            }

            for data_type, pipeline in to_check.items():
                for sub_name, sub_step in pipeline.items():
                    # If it is a Choice, make sure it is the correct one!
                    if isinstance(sub_step, AutoSklearnChoice):
                        key = "data_preprocessor:feature_type:{}:{}:__choice__".format(
                            data_type, sub_name
                        )
                        keys_checked.extend(
                            self._test_set_hyperparameter_choice(
                                key, sub_step, config_dict
                            )
                        )
                    # If it is a component, make sure it has the correct hyperparams
                    elif isinstance(sub_step, AutoSklearnComponent):
                        keys_checked.extend(
                            self._test_set_hyperparameter_component(
                                "data_preprocessor:feature_type:{}:{}".format(
                                    data_type, sub_name
                                ),
                                sub_step,
                                config_dict,
                            )
                        )
                    else:
                        raise ValueError("New type of pipeline component!")
            return keys_checked

        for key, value in config_dict.items():
            if key != expected_key and expected_sub_key in key:
                expected_attributes[key.split(":")[-1]] = value
                keys_checked.append(key)

        if expected_attributes:
            attributes = vars(implementation.choice)
            # Cannot check the whole dictionary, just names, as some
            # classes map the text hyperparameter directly to a function!
            for expected_attribute in expected_attributes.keys():
                self.assertIn(expected_attribute, attributes.keys())

        return keys_checked

    def _test_set_hyperparameter_component(
        self, expected_key, implementation, config_dict
    ):
        """
        Given a configuration in config, this procedure makes sure that
        the given implementation, which should be a autosklearn component, honors
        is created with the desired hyperparameters stated in config_dict

        TODO: typing
        """
        keys_checked = []
        attributes = vars(implementation)
        expected_attributes = {}
        for key, value in config_dict.items():
            if expected_key in key:
                keys_checked.append(key)
                key = key.replace(expected_key + ":", "")
                if ":" in key:
                    raise ValueError(
                        "This utility should only be called with a "
                        "matching string that produces leaf configurations, "
                        "that is no further colons are expected, yet key={}"
                        "".format(key)
                    )
                expected_attributes[key] = value
        # self.assertDictContainsSubset(expected_attributes, attributes)
        # Cannot check the whole dictionary, just names, as some
        # classes map the text hyperparameter directly to a function!
        for expected_attribute in expected_attributes.keys():
            self.assertIn(expected_attribute, attributes.keys())
        return keys_checked

    def test_set_hyperparameters_honors_configuration(self):
        """Makes sure that a given configuration is honored in practice.

        This method tests that the set hyperparameters actually create objects
        that comply with the given configuration. It iterates trough the pipeline to
        make sure we did not miss a step, but also checks at the end that every
        configuration from Config was checked.

        Also considers random_state and ensures pipeline steps correctly recieve
        the right random_state
        """
        random_state = 1
        all_combinations = list(itertools.product([True, False], repeat=4))
        for (
            sparse,
            multilabel,
            signed,
            multiclass,
        ) in all_combinations:
            dataset_properties = {
                "sparse": sparse,
                "multilabel": multilabel,
                "multiclass": multiclass,
                "signed": signed,
            }
            cls = SimpleClassificationPipeline(
                random_state=random_state,
                dataset_properties=dataset_properties,
            )
            cs = cls.get_hyperparameter_search_space()
            config = cs.sample_configuration()

            # Set hyperparameters takes a given config and translate
            # a config to an actual implementation
            cls.set_hyperparameters(config)
            config_dict = config.get_dictionary()

            # keys to check is our mechanism to ensure that every
            # every config key is checked
            keys_checked = []

            for name, step in cls.named_steps.items():
                if name == "data_preprocessor":
                    keys_checked.extend(
                        self._test_set_hyperparameter_choice(
                            "data_preprocessor:__choice__", step, config_dict
                        )
                    )
                    self.assertEqual(step.random_state, random_state)
                elif name == "balancing":
                    keys_checked.extend(
                        self._test_set_hyperparameter_component(
                            "balancing", step, config_dict
                        )
                    )
                elif name == "feature_preprocessor":
                    keys_checked.extend(
                        self._test_set_hyperparameter_choice(
                            "feature_preprocessor:__choice__", step, config_dict
                        )
                    )
                    self.assertEqual(step.random_state, random_state)
                elif name == "classifier":
                    keys_checked.extend(
                        self._test_set_hyperparameter_choice(
                            "classifier:__choice__", step, config_dict
                        )
                    )
                    self.assertEqual(step.random_state, random_state)
                else:
                    raise ValueError(
                        "Found another type of step! Need to update this check"
                    )

            # Make sure we checked the whole configuration
            self.assertSetEqual(set(config_dict.keys()), set(keys_checked))

    def test_fit_instantiates_component(self):
        """Make sure that if a preprocessor is added, it's fit
        method is called"""
        preprocessing_components.add_preprocessor(CrashPreprocessor)

        # We reduce the search space as forbidden clauses prevent to instantiate
        # the user defined preprocessor manually
        cls = SimpleClassificationPipeline(include={"classifier": ["random_forest"]})
        cs = cls.get_hyperparameter_search_space()
        self.assertIn("CrashPreprocessor", str(cs))
        config = cs.sample_configuration()
        try:
            config["feature_preprocessor:__choice__"] = "CrashPreprocessor"
        except Exception as e:
            # In case of failure clean up the components and print enough information
            # to clean up with check in the future
            del preprocessing_components.additional_components.components[
                "CrashPreprocessor"
            ]
            self.fail("cs={} config={} Exception={}".format(cs, config, e))

        cls.set_hyperparameters(config)

        with self.assertRaisesRegex(ValueError, "Make sure fit is called"):
            with ignore_warnings(classifier_warnings):
                cls.fit(
                    X=np.array([[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12]]),
                    y=np.array([1, 0, 1, 1]),
                )

        del preprocessing_components.additional_components.components[
            "CrashPreprocessor"
        ]

    def test_get_hyperparameter_search_space_feat_type(self):
        cs_mc = SimpleClassificationPipeline(
            feat_type={"1": "numerical"}
        ).get_hyperparameter_search_space(dataset_properties={"multiclass": True})
        self.assertNotIn("data_preprocessor:feature_type:categorical", str(cs_mc))
        self.assertNotIn("data_preprocessor:feature_type:text", str(cs_mc))

        cs_mc = SimpleClassificationPipeline(
            feat_type={"1": "categorical"}
        ).get_hyperparameter_search_space(dataset_properties={"multilabel": True})
        self.assertNotIn("data_preprocessor:feature_type:numerical", str(cs_mc))
        self.assertNotIn("data_preprocessor:feature_type:text", str(cs_mc))

        cs_mc = SimpleClassificationPipeline(
            feat_type={"1": "string"}
        ).get_hyperparameter_search_space(dataset_properties={"sparse": True})
        self.assertNotIn("data_preprocessor:feature_type:numerical", str(cs_mc))
        self.assertNotIn("data_preprocessor:feature_type:categorical", str(cs_mc))

        cs_mc = SimpleClassificationPipeline(
            feat_type={"1": "numerical", "2": "categorical"}
        ).get_hyperparameter_search_space(
            dataset_properties={"multilabel": True, "multiclass": True}
        )
        self.assertNotIn("data_preprocessor:feature_type:text", str(cs_mc))

        cs_mc = SimpleClassificationPipeline(
            feat_type={"1": "numerical", "2": "string"}
        ).get_hyperparameter_search_space(
            dataset_properties={"multilabel": True, "multiclass": True}
        )
        self.assertNotIn("data_preprocessor:feature_type:categorical", str(cs_mc))

        cs_mc = SimpleClassificationPipeline(
            feat_type={"1": "categorical", "2": "string"}
        ).get_hyperparameter_search_space(
            dataset_properties={"multilabel": True, "multiclass": True}
        )
        self.assertNotIn("data_preprocessor:feature_type:numerical", str(cs_mc))



from typing import Any

import numpy as np

from autosklearn.metrics import accuracy, make_scorer


def _accuracy_requiring_X_data(
    y_true: np.ndarray,
    y_pred: np.ndarray,
    X_data: Any,
) -> float:
    """Dummy metric that needs X Data"""
    if X_data is None:
        raise ValueError()
    return accuracy(y_true, y_pred)


acc_with_X_data = make_scorer(
    name="acc_with_X_data",
    score_func=_accuracy_requiring_X_data,
    needs_X=True,
    optimum=1,
    worst_possible_result=0,
    greater_is_better=True,
)



from pathlib import Path

from dask.distributed import Client, LocalCluster

from autosklearn.util.dask import LocalDask, UserDask

import pytest


@pytest.mark.parametrize("n_jobs", [1, 2])
def test_user_dask(tmp_path: Path, n_jobs: int) -> None:
    """
    Expects
    -------
    * A UserDask should not close the client after exiting context
    """
    cluster = LocalCluster(
        n_workers=n_jobs,
        processes=False,
        threads_per_worker=1,
        local_directory=tmp_path,
    )
    client = Client(cluster, heartbeat_interval=10000)

    # Active at creation
    dask = UserDask(client)

    client_1 = None
    with dask as user_client:
        client_1 = user_client
        assert user_client.status == "running"

    client_2 = None
    with dask as user_client:
        assert user_client.status == "running"
        client_2 = user_client

    # Make sure they are the same client
    assert id(client_1) == id(client_2)

    # Remains running after context
    assert client_1.status == "running"

    cluster.close()
    client.close()

    assert client.status == "closed"


def test_local_dask_creates_new_clients(tmp_path: Path) -> None:
    """
    Expects
    -------
    * A LocalDask should create new dask clusters at each context usage
    """
    # We need 2 to use an actual dask client and not a SingleThreadedClient
    local_dask = LocalDask(n_jobs=2)

    client_1 = None
    with local_dask as client:
        client_1 = client
        assert client_1.status == "running"

    assert client_1.status == "closed"

    client_2 = None
    with local_dask as client:
        client_2 = client
        assert client_2.status == "running"

    # Make sure they were different clients
    assert id(client_1) != id(client_2)

    assert client_2.status == "closed"
    assert client_1.status == "closed"



import numpy as np
import scipy.sparse
import sklearn.datasets
import sklearn.model_selection
import sklearn.pipeline
import sklearn.tree
from sklearn.impute import SimpleImputer
from sklearn.tree import DecisionTreeClassifier

from autosklearn.pipeline.implementations.CategoryShift import CategoryShift
from autosklearn.pipeline.implementations.SparseOneHotEncoder import SparseOneHotEncoder

import unittest

sparse1 = scipy.sparse.csc_matrix(
    ([3, 2, 1, 1, 2, 3], ((1, 4, 5, 2, 3, 5), (0, 0, 0, 1, 1, 1))), shape=(6, 2)
)
sparse1_1h = scipy.sparse.csc_matrix(
    ([1, 1, 1, 1, 1, 1], ((5, 4, 1, 2, 3, 5), (0, 1, 2, 3, 4, 5))), shape=(6, 6)
)

sparse2 = scipy.sparse.csc_matrix(
    ([2, 1, 0, 0, 0, 0], ((1, 4, 5, 2, 3, 5), (0, 0, 0, 1, 1, 1))), shape=(6, 2)
)
sparse2_1h = scipy.sparse.csc_matrix(
    ([1, 1, 1, 1, 1, 1], ((5, 4, 1, 2, 3, 5), (0, 1, 2, 3, 3, 3))), shape=(6, 4)
)

sparse2_csr = scipy.sparse.csr_matrix(
    ([2, 1, 0, 0, 0, 0], ((1, 4, 5, 2, 3, 5), (0, 0, 0, 1, 1, 1))), shape=(6, 2)
)
sparse2_csr_1h = scipy.sparse.csr_matrix(
    ([1, 1, 1, 1, 1, 1], ((5, 4, 1, 2, 3, 5), (0, 1, 2, 3, 3, 3))), shape=(6, 4)
)


class TestSparseOneHotEncoder(unittest.TestCase):
    def test_sparse1(self):
        self._fit_then_transform(sparse1_1h.todense(), sparse1)

    def test_sparse2(self):
        self._fit_then_transform(sparse2_1h.todense(), sparse2)

    def test_sparse2_csr(self):
        self._fit_then_transform(sparse2_csr_1h.todense(), sparse2_csr)

    def _fit_then_transform(self, expected, input):
        # Test fit_transform
        input_copy = input.copy()
        ohe = SparseOneHotEncoder()
        transformation = ohe.fit_transform(input)
        self.assertIsInstance(transformation, scipy.sparse.csr_matrix)
        np.testing.assert_array_almost_equal(
            expected.astype(float), transformation.todense()
        )
        self._check_arrays_equal(input, input_copy)

        # Test fit, and afterwards transform
        ohe2 = SparseOneHotEncoder()
        ohe2.fit(input)
        transformation = ohe2.transform(input)
        self.assertIsInstance(transformation, scipy.sparse.csr_matrix)
        np.testing.assert_array_almost_equal(expected, transformation.todense())
        self._check_arrays_equal(input, input_copy)

    def _check_arrays_equal(self, a1, a2):
        if scipy.sparse.issparse(a1):
            a1 = a1.toarray()
        if scipy.sparse.issparse(a2):
            a2 = a2.toarray()
        np.testing.assert_array_almost_equal(a1, a2)

    def test_transform_with_unknown_value(self):
        # fit_data: this is going to be used to fit.
        # note that 0 is no category because the data here is sparse.
        fit_data = np.array(((0, 1, 2, 3, 4, 5), (0, 1, 2, 3, 4, 5))).transpose()
        fds = scipy.sparse.csr_matrix(fit_data)
        ohe = SparseOneHotEncoder()
        ohe.fit(fds)
        # transf_data: this is going to be used in a transform call.
        # Note that transf_data has categories not seen at the fit.
        # Unseen categories are ignored (encoded just with zeros).
        transf_data = np.array(((0, 1, 2, 6), (0, 1, 6, 7))).transpose()
        tds = scipy.sparse.csr_matrix(transf_data)
        output = ohe.transform(tds).todense()
        # From tds, just 3 categories (1 and 2 in the 1st feature and 1 in the 2nd
        # feature) have been seen during fit, therefore:
        self.assertEqual(3, np.sum(output))

    def test_classification_workflow(self):
        X, y = sklearn.datasets.fetch_openml(
            data_id=24, as_frame=False, return_X_y=True
        )
        print(type(X))

        X_train, X_test, y_train, y_test = sklearn.model_selection.train_test_split(
            X, y, random_state=3, train_size=0.5, test_size=0.5
        )

        X_train = scipy.sparse.csc_matrix(X_train)
        X_test = scipy.sparse.csc_matrix(X_test)

        pipeline = sklearn.pipeline.Pipeline(
            (
                ("shift", CategoryShift()),
                ("imput", SimpleImputer(strategy="constant", fill_value=2)),
                ("ohe", SparseOneHotEncoder()),
                ("tree", DecisionTreeClassifier(random_state=1)),
            )
        )

        pipeline.fit(X_train, y_train)
        pred_train = pipeline.predict(X_train)
        self.assertTrue((pred_train == y_train).all())
        # With an incorrect copy operation the OneHotEncoder would rearrange
        # the data in such a way that the accuracy would drop to 66%
        pred_test = pipeline.predict(X_test)
        self.assertTrue((pred_test == y_test).all())



from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
from ConfigSpace.configuration_space import ConfigurationSpace
from ConfigSpace.hyperparameters import CategoricalHyperparameter
from sklearn.base import BaseEstimator

from autosklearn.askl_typing import FEAT_TYPE_TYPE
from autosklearn.pipeline.base import DATASET_PROPERTIES_TYPE, PIPELINE_DATA_DTYPE
from autosklearn.pipeline.components.base import AutoSklearnPreprocessingAlgorithm
from autosklearn.pipeline.constants import (
    DENSE,
    INPUT,
    SIGNED_DATA,
    SPARSE,
    UNSIGNED_DATA,
)


class Balancing(AutoSklearnPreprocessingAlgorithm):
    def __init__(
        self,
        strategy: str = "none",
        random_state: Optional[Union[int, np.random.RandomState]] = None,
    ) -> None:
        self.strategy = strategy
        self.random_state = random_state

    def fit(
        self, X: PIPELINE_DATA_DTYPE, y: Optional[PIPELINE_DATA_DTYPE] = None
    ) -> "Balancing":
        self.fitted_ = True
        return self

    def transform(self, X: PIPELINE_DATA_DTYPE) -> PIPELINE_DATA_DTYPE:
        return X

    def get_weights(
        self,
        Y: PIPELINE_DATA_DTYPE,
        classifier: BaseEstimator,
        preprocessor: BaseEstimator,
        init_params: Optional[Dict[str, Any]],
        fit_params: Optional[Dict[str, Any]],
    ) -> Tuple[Optional[Dict[str, Any]], Optional[Dict[str, Any]]]:
        if init_params is None:
            init_params = {}

        if fit_params is None:
            fit_params = {}

        # Classifiers which require sample weights:
        # We can have adaboost in here, because in the fit method,
        # the sample weights are normalized:
        # https://github.com/scikit-learn/scikit-learn/blob/0.15.X/sklearn/ensemble/weight_boosting.py#L121
        # Have RF and ET in here because they emit a warning if class_weights
        #  are used together with warmstarts
        clf_ = [
            "adaboost",
            "random_forest",
            "extra_trees",
            "sgd",
            "passive_aggressive",
            "gradient_boosting",
        ]
        pre_: List[str] = []
        if classifier in clf_ or preprocessor in pre_:
            if len(Y.shape) > 1:
                offsets = [2**i for i in range(Y.shape[1])]
                Y_ = np.sum(Y * offsets, axis=1)
            else:
                Y_ = Y

            unique, counts = np.unique(Y_, return_counts=True)
            # This will result in an average weight of 1!
            cw = 1 / (counts / np.sum(counts)) / 2
            if len(Y.shape) == 2:
                cw /= Y.shape[1]

            sample_weights = np.ones(Y_.shape)

            for i, ue in enumerate(unique):
                mask = Y_ == ue
                sample_weights[mask] *= cw[i]

            if classifier in clf_:
                fit_params["classifier:sample_weight"] = sample_weights
            if preprocessor in pre_:
                fit_params["feature_preprocessor:sample_weight"] = sample_weights

        # Classifiers which can adjust sample weights themselves via the
        # argument `class_weight`
        clf_ = ["decision_tree", "liblinear_svc", "libsvm_svc"]
        pre_ = ["liblinear_svc_preprocessor", "extra_trees_preproc_for_classification"]
        if classifier in clf_:
            init_params["classifier:class_weight"] = "balanced"
        if preprocessor in pre_:
            init_params["feature_preprocessor:class_weight"] = "balanced"

        clf_ = ["ridge"]
        if classifier in clf_:
            class_weights = {}

            unique, counts = np.unique(Y, return_counts=True)
            cw = 1.0 / counts
            cw = cw / np.mean(cw)

            for i, ue in enumerate(unique):
                class_weights[ue] = cw[i]

            if classifier in clf_:
                init_params["classifier:class_weight"] = class_weights

        return init_params, fit_params

    @staticmethod
    def get_properties(
        dataset_properties: Optional[DATASET_PROPERTIES_TYPE] = None,
    ) -> Dict[str, Optional[Union[str, int, bool, Tuple]]]:
        return {
            "shortname": "Balancing",
            "name": "Balancing Imbalanced Class Distributions",
            "handles_missing_values": True,
            "handles_nominal_values": True,
            "handles_numerical_features": True,
            "prefers_data_scaled": False,
            "prefers_data_normalized": False,
            "handles_regression": False,
            "handles_classification": True,
            "handles_multiclass": True,
            "handles_multilabel": True,
            "handles_multioutput": False,
            "is_deterministic": True,
            "handles_sparse": True,
            "handles_dense": True,
            "input": (DENSE, SPARSE, UNSIGNED_DATA, SIGNED_DATA),
            "output": (INPUT,),
            "preferred_dtype": None,
        }

    @staticmethod
    def get_hyperparameter_search_space(
        feat_type: Optional[FEAT_TYPE_TYPE] = None,
        dataset_properties: Optional[DATASET_PROPERTIES_TYPE] = None,
    ) -> ConfigurationSpace:
        # TODO add replace by zero!
        strategy = CategoricalHyperparameter(
            "strategy", ["none", "weighting"], default_value="none"
        )
        cs = ConfigurationSpace()
        cs.add_hyperparameter(strategy)
        return cs



import sklearn.ensemble

from autosklearn.pipeline.components.classification.random_forest import RandomForest

from .test_base import BaseClassificationComponentTest


class RandomForestComponentTest(BaseClassificationComponentTest):

    __test__ = True

    res = dict()
    res["default_iris"] = 0.96
    res["iris_n_calls"] = 9
    res["default_iris_iterative"] = res["default_iris"]
    res["default_iris_proba"] = 0.0996785324703419
    res["default_iris_sparse"] = 0.85999999999999999
    res["default_digits"] = 0.8998178506375227
    res["digits_n_calls"] = 9
    res["default_digits_iterative"] = res["default_digits"]
    res["default_digits_binary"] = 0.9896782027929569
    res["default_digits_multilabel"] = 0.9973653110879388
    res["default_digits_multilabel_proba"] = 0.9965660960196189

    sk_mod = sklearn.ensemble.RandomForestClassifier
    module = RandomForest
    step_hyperparameter = {
        "name": "n_estimators",
        "value": module.get_max_iter(),
    }



""" Provides simplified 2 use cases of dask that we consider

1. A UserDask is when a user supplies a dask client, in which case
we don't close this down and leave it up to the user to control its lifetime.
2.  A LocalDask is one we use when no user dask is supplied. In this case
we make sure to spin up and close down clients as needed.

Both of these can be uniformly accessed as a context manager.

.. code:: python

    # Locally controlled dask client
    local_dask = LocalDask(n_jobs=2)
    with local_dask as client:
        # Do stuff with client
        ...

    # `client` is shutdown properly

    # ----------------

    # User controlled dask client
    user_dask = UserDask(user_client)

    with user_dask as client:
        # Do stuff with (client == user_client)
        ...

    # `user_client` is still open and up to the user to close
"""
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any

import tempfile

from dask.distributed import Client, LocalCluster

from autosklearn.util.single_thread_client import SingleThreadedClient


class Dask(ABC):
    @abstractmethod
    def client(self) -> Client:
        """Should return a dask client"""
        ...

    @abstractmethod
    def close(self) -> None:
        """Should close up any resources needed for the dask client"""
        ...

    def __enter__(self) -> Client:
        return self.client()

    def __exit__(self, *args: Any, **kwargs: Any) -> None:
        self.close()

    @abstractmethod
    def __repr__(self) -> str:
        ...


class UserDask(Dask):
    """A dask instance created by a user"""

    def __init__(self, client: Client):
        """
        Parameters
        ----------
        client : Client
            The client they passed in
        """
        self._client = client

    def client(self) -> Client:
        """The dask client"""
        return self._client

    def close(self) -> None:
        """Close the dask client"""
        # We do nothing, it's user provided
        pass

    def __repr__(self) -> str:
        return "UserDask(...)"


class LocalDask(Dask):
    def __init__(self, n_jobs: int | None = None) -> None:
        self.n_jobs = n_jobs
        self._client: Client | None = None
        self._cluster: LocalCluster | None = None

    def client(self) -> Client:
        """Creates a usable dask client or returns an existing one

        If there is not current client, because it has been closed, create
        a new one.
        * If ``n_jobs == 1``, create a ``SingleThreadedClient``
        * Else create a ``Client`` with a ``LocalCluster``
        """
        if self._client is not None:
            return self._client

        if self.n_jobs == 1:
            cluster = None
            client = SingleThreadedClient()
        else:
            cluster = LocalCluster(
                n_workers=self.n_jobs,
                processes=False,
                threads_per_worker=1,
                # We use tmpdir to save the workers as deleting workers takes
                # more time than deleting backend directories.
                # This prevent an error saying that the worker file was deleted,
                # so the client could not close the worker properly
                local_directory=tempfile.gettempdir(),
                # Memory is handled by the pynisher, not by the dask worker/nanny
                memory_limit=0,
            )
            client = Client(cluster, heartbeat_interval=10000)  # 10s

        self._client = client
        self._cluster = cluster
        return self._client

    def close(self) -> None:
        """Closes any open dask client"""
        if self._client is None:
            return

        self._client.close()
        if self._cluster is not None:
            self._cluster.close()

        self._client = None
        self._cluster = None

    def __repr__(self) -> str:
        return f"LocalDask(n_jobs = {self.n_jobs})"



import numpy as np

from autosklearn.pipeline.components.feature_preprocessing.densifier import Densifier
from autosklearn.pipeline.util import PreprocessingTestCase, _test_preprocessing


class DensifierComponentTest(PreprocessingTestCase):
    def test_default_configuration(self):
        transformation, original = _test_preprocessing(Densifier, make_sparse=True)
        self.assertIsInstance(transformation, np.ndarray)
        self.assertEqual(transformation.shape, original.shape)
        self.assertIsInstance(transformation, np.ndarray)

    def test_preprocessing_dtype(self):
        super(DensifierComponentTest, self)._test_preprocessing_dtype(Densifier)



from __future__ import annotations

from typing import Tuple

import re
from pathlib import Path

import numpy as np

from autosklearn.util.disk import sizeof

RunID = Tuple[int, int, float]


class Run:
    """Class for storing information about a run used during ensemble building.

    Note
    ----
    This is for internal use by the EnsembleBuilder and not for general usage.
    """

    # For matching prediction files
    RE_MODEL_PREDICTION_FILE = (
        r"^predictions_ensemble_([0-9]*)_([0-9]*)_([0-9]{1,3}\.[0-9]*).npy$"
    )

    # For matching run directories
    RE_MODEL_DIR = r"^([0-9]*)_([0-9]*)_([0-9]{1,3}\.[0-9]*)$"

    def __init__(self, path: Path) -> None:
        """Creates a Run from a path pointing to the directory of a run

        Parameters
        ----------
        path: Path
            Expects something like /path/to/{seed}_{numrun}_{budget}
        """
        name = path.name
        seed, num_run, budget = name.split("_")

        self.dir = path
        self.seed = int(seed)
        self.num_run = int(num_run)
        self.budget = float(budget)

        # These are ordered based on preference
        self.losses: dict[str, float] = {}

        self._mem_usage: float | None = None

        # Items that will be delete when the run is saved back to file
        self._cache: dict[str, np.ndarray] = {}

        # The recorded time of ensemble/test predictions modified
        self.recorded_mtimes: dict[str, float] = {}
        self.record_modified_times()

    @property
    def mem_usage(self) -> float:
        """The memory usage of this run based on it's directory"""
        if self._mem_usage is None:
            self._mem_usage = round(sizeof(self.dir, unit="MB"), 2)

        return self._mem_usage

    def is_dummy(self) -> bool:
        """Whether this run is a dummy run or not"""
        return self.num_run == 1

    def was_modified(self) -> bool:
        """Query for when the ens file was last modified"""
        recorded = self.recorded_mtimes.get("ensemble")
        last = self.pred_path().stat().st_mtime
        return recorded != last

    def pred_path(self, kind: str = "ensemble") -> Path:
        """Get the path to certain predictions"""
        fname = f"predictions_{kind}_{self.seed}_{self.num_run}_{self.budget}.npy"
        return self.dir / fname

    def record_modified_times(self) -> None:
        """Records the last time each prediction file type was modified, if it exists"""
        self.recorded_mtimes = {}
        for kind in ["ensemble", "test"]:
            path = self.pred_path(kind)  # type: ignore
            if path.exists():
                self.recorded_mtimes[kind] = path.stat().st_mtime

    def has_predictions(self, kind: str = "ensemble") -> bool:
        """
        Parameters
        ----------
        kind: "ensemble" | "test" = "ensemble"
            The kind of predictions to query for

        Returns
        -------
        bool
            Whether this run has the kind of predictions queried for
        """
        return self.pred_path(kind).exists()

    def predictions(
        self,
        kind: str = "ensemble",
        precision: int | None = None,
    ) -> np.ndarray:
        """Load the predictions for this run

        Parameters
        ----------
        kind : "ensemble" | "test"
            The kind of predictions to load

        precisions : type | None = None
            What kind of precision reduction to apply

        Returns
        -------
        np.ndarray
            The loaded predictions
        """
        key = f"predictions_{kind}"
        if key in self._cache:
            return self._cache[key]

        path = self.pred_path(kind)

        with path.open("rb") as f:
            # TODO: We should probably remove this requirement. I'm not sure why model
            # predictions are being saved as pickled
            predictions = np.load(f, allow_pickle=True)

        if precision:
            dtypes: dict[int, type] = {16: np.float16, 32: np.float32, 64: np.float64}
            dtype = dtypes.get(precision, None)

            if dtype is not None:
                predictions = predictions.astype(dtype=dtype, copy=False)

        self._cache[key] = predictions
        return predictions

    def __getstate__(self) -> dict:
        """Remove the cache when pickling."""
        state = self.__dict__.copy()
        del state["_cache"]
        return state

    def __setstate__(self, state: dict) -> None:
        """Reset state and instansiate blank cache."""
        self.__dict__.update(state)
        self._cache = {}

    @property
    def id(self) -> RunID:
        """Get the three components of it's id"""
        return self.seed, self.num_run, self.budget

    def __hash__(self) -> int:
        return hash(self.id)

    def __repr__(self) -> str:
        return f"Run(id={self.id}, losses={self.losses})"

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Run) and other.id == self.id

    @staticmethod
    def valid(path: Path) -> bool:
        """
        Parameters
        ----------
        path: Path
            The path to check

        Returns
        -------
        bool
            Whether the path is a valid run dir
        """
        return re.match(Run.RE_MODEL_DIR, path.name) is not None



import sklearn.ensemble

from autosklearn.pipeline.components.regression.gradient_boosting import (
    GradientBoosting,
)

from .test_base import BaseRegressionComponentTest


class GradientBoostingComponentTest(BaseRegressionComponentTest):

    __test__ = True

    res = dict()
    res["default_boston"] = 0.7491382574462079
    res["default_boston_iterative"] = 0.7491382574462079
    res["default_boston_sparse"] = None
    res["boston_n_calls"] = 9
    res["default_diabetes"] = 0.2872735632261877
    res["default_diabetes_iterative"] = 0.2872735632261877
    res["default_diabetes_sparse"] = None
    res["diabetes_n_call"] = 11

    sk_mod = sklearn.ensemble.GradientBoostingRegressor
    module = GradientBoosting



import copy
from collections import OrderedDict, defaultdict, deque

import numpy as np
import pandas as pd
import scipy.sparse
import scipy.stats
from scipy.linalg import LinAlgError

# TODO use balanced accuracy!
from sklearn.multiclass import OneVsRestClassifier
from sklearn.utils import check_array
from sklearn.utils.multiclass import type_of_target

from autosklearn.pipeline.components.data_preprocessing.feature_type import (
    FeatTypeSplit,
)

from .metafeature import DatasetMetafeatures, HelperFunction, MetaFeature


# TODO Allow multiple dependencies for a metafeature
# TODO Add HelperFunction as an object
class HelperFunctions(object):
    def __init__(self):
        self.functions = OrderedDict()
        self.values = OrderedDict()

    def clear(self):
        self.values = OrderedDict()
        self.computation_time = OrderedDict()

    def __iter__(self):
        return self.functions.__iter__()

    def __getitem__(self, item):
        return self.functions.__getitem__(item)

    def __setitem__(self, key, value):
        return self.functions.__setitem__(key, value)

    def __delitem__(self, key):
        return self.functions.__delitem__(key)

    def __contains__(self, item):
        return self.functions.__contains__(item)

    def is_calculated(self, key):
        """Return if a helper function has already been executed.

        Necessary as get_value() can return None if the helper function hasn't
        been executed or if it returned None."""
        return key in self.values

    def get_value(self, key):
        return self.values.get(key).value

    def set_value(self, key, item):
        self.values[key] = item

    def define(self, name):
        """Decorator for adding helper functions to a "dictionary".
        This behaves like a function decorating a function,
        not a class decorating a function"""

        def wrapper(metafeature_class):
            instance = metafeature_class()
            self.__setitem__(name, instance)
            return instance

        return wrapper


class MetafeatureFunctions(object):
    def __init__(self):
        self.functions = OrderedDict()
        self.dependencies = OrderedDict()
        self.values = OrderedDict()

    def clear(self):
        self.values = OrderedDict()

    def __iter__(self):
        return self.functions.__iter__()

    def __getitem__(self, item):
        return self.functions.__getitem__(item)

    def __setitem__(self, key, value):
        return self.functions.__setitem__(key, value)

    def __delitem__(self, key):
        return self.functions.__delitem__(key)

    def __contains__(self, item):
        return self.functions.__contains__(item)

    def get_value(self, key):
        return self.values[key].value

    def set_value(self, key, item):
        self.values[key] = item

    def is_calculated(self, key):
        """Return if a helper function has already been executed.

        Necessary as get_value() can return None if the helper function hasn't
        been executed or if it returned None."""
        return key in self.values

    def get_dependency(self, name):
        """Return the dependency of metafeature "name"."""
        return self.dependencies.get(name)

    def define(self, name, dependency=None):
        """Decorator for adding metafeature functions to a "dictionary" of
        metafeatures. This behaves like a function decorating a function,
        not a class decorating a function"""

        def wrapper(metafeature_class):
            instance = metafeature_class()
            self.__setitem__(name, instance)
            self.dependencies[name] = dependency
            return instance

        return wrapper


metafeatures = MetafeatureFunctions()
helper_functions = HelperFunctions()


################################################################################
# Simple features
################################################################################
@metafeatures.define("NumberOfInstances")
class NumberOfInstances(MetaFeature):
    def _calculate(self, X, y, logger, feat_type):
        return float(X.shape[0])


@metafeatures.define("LogNumberOfInstances", dependency="NumberOfInstances")
class LogNumberOfInstances(MetaFeature):
    def _calculate(self, X, y, logger, feat_type):
        return np.log(metafeatures.get_value("NumberOfInstances"))


@metafeatures.define("NumberOfClasses")
class NumberOfClasses(MetaFeature):
    """
    Calculate the number of classes.

    Calls np.unique on the targets. If the dataset is a multilabel dataset,
    does this for each label seperately and returns the mean.
    """

    def _calculate(self, X, y, logger, feat_type):
        if type_of_target(y) == "multilabel-indicator":
            # We have a label binary indicator array:
            # each sample is one row of a 2d array of shape (n_samples, n_classes)
            return y.shape[1]
        if len(y.shape) == 2:
            return np.mean([len(np.unique(y[:, i])) for i in range(y.shape[1])])
        else:
            return float(len(np.unique(y)))


@metafeatures.define("NumberOfFeatures")
class NumberOfFeatures(MetaFeature):
    def _calculate(self, X, y, logger, feat_type):
        return float(X.shape[1])


@metafeatures.define("LogNumberOfFeatures", dependency="NumberOfFeatures")
class LogNumberOfFeatures(MetaFeature):
    def _calculate(self, X, y, logger, feat_type):
        return np.log(metafeatures.get_value("NumberOfFeatures"))


@helper_functions.define("MissingValues")
class MissingValues(HelperFunction):
    def _calculate(self, X, y, logger, feat_type):
        missing = pd.isna(X)
        return missing

    def _calculate_sparse(self, X, y, logger, feat_type):
        data = [True if not np.isfinite(x) else False for x in X.data]
        missing = X.__class__((data, X.indices, X.indptr), shape=X.shape, dtype=bool)
        return missing


@metafeatures.define("NumberOfInstancesWithMissingValues", dependency="MissingValues")
class NumberOfInstancesWithMissingValues(MetaFeature):
    def _calculate(self, X, y, logger, feat_type):
        missing = helper_functions.get_value("MissingValues")
        num_missing = missing.sum(axis=1)
        return float(np.sum([1 if num > 0 else 0 for num in num_missing]))

    def _calculate_sparse(self, X, y, logger, feat_type):
        missing = helper_functions.get_value("MissingValues")
        new_missing = missing.tocsr()
        num_missing = [
            np.sum(new_missing.data[new_missing.indptr[i] : new_missing.indptr[i + 1]])
            for i in range(new_missing.shape[0])
        ]

        return float(np.sum([1 if num > 0 else 0 for num in num_missing]))


@metafeatures.define(
    "PercentageOfInstancesWithMissingValues",
    dependency="NumberOfInstancesWithMissingValues",
)
class PercentageOfInstancesWithMissingValues(MetaFeature):
    def _calculate(self, X, y, logger, feat_type):
        n_missing = metafeatures.get_value("NumberOfInstancesWithMissingValues")
        n_total = float(metafeatures["NumberOfInstances"](X, y, logger).value)
        return float(n_missing / n_total)


@metafeatures.define("NumberOfFeaturesWithMissingValues", dependency="MissingValues")
class NumberOfFeaturesWithMissingValues(MetaFeature):
    def _calculate(self, X, y, logger, feat_type):
        missing = helper_functions.get_value("MissingValues")
        num_missing = missing.sum(axis=0)
        return float(np.sum([1 if num > 0 else 0 for num in num_missing]))

    def _calculate_sparse(self, X, y, logger, feat_type):
        missing = helper_functions.get_value("MissingValues")
        new_missing = missing.tocsc()
        num_missing = [
            np.sum(new_missing.data[new_missing.indptr[i] : new_missing.indptr[i + 1]])
            for i in range(missing.shape[1])
        ]

        return float(np.sum([1 if num > 0 else 0 for num in num_missing]))


@metafeatures.define(
    "PercentageOfFeaturesWithMissingValues",
    dependency="NumberOfFeaturesWithMissingValues",
)
class PercentageOfFeaturesWithMissingValues(MetaFeature):
    def _calculate(self, X, y, logger, feat_type):
        n_missing = metafeatures.get_value("NumberOfFeaturesWithMissingValues")
        n_total = float(metafeatures["NumberOfFeatures"](X, y, logger).value)
        return float(n_missing / n_total)


@metafeatures.define("NumberOfMissingValues", dependency="MissingValues")
class NumberOfMissingValues(MetaFeature):
    def _calculate(self, X, y, logger, feat_type):
        if scipy.sparse.issparse(X):
            return float(helper_functions.get_value("MissingValues").sum())
        else:
            return float(np.count_nonzero(helper_functions.get_value("MissingValues")))


@metafeatures.define("PercentageOfMissingValues", dependency="NumberOfMissingValues")
class PercentageOfMissingValues(MetaFeature):
    def _calculate(self, X, y, logger, feat_type):
        return float(metafeatures.get_value("NumberOfMissingValues")) / float(
            X.shape[0] * X.shape[1]
        )


# TODO: generalize this!
@metafeatures.define("NumberOfNumericFeatures")
class NumberOfNumericFeatures(MetaFeature):
    def _calculate(self, X, y, logger, feat_type):
        return np.sum([value == "numerical" for value in feat_type.values()])


@metafeatures.define("NumberOfCategoricalFeatures")
class NumberOfCategoricalFeatures(MetaFeature):
    def _calculate(self, X, y, logger, feat_type):
        return np.sum([value == "categorical" for value in feat_type.values()])


@metafeatures.define("RatioNumericalToNominal")
class RatioNumericalToNominal(MetaFeature):
    def _calculate(self, X, y, logger, feat_type):
        num_categorical = float(
            metafeatures["NumberOfCategoricalFeatures"](X, y, logger, feat_type).value
        )
        num_numerical = float(
            metafeatures["NumberOfNumericFeatures"](X, y, logger, feat_type).value
        )
        if num_categorical == 0.0:
            return 0.0
        return num_numerical / num_categorical


@metafeatures.define("RatioNominalToNumerical")
class RatioNominalToNumerical(MetaFeature):
    def _calculate(self, X, y, logger, feat_type):
        num_categorical = float(
            metafeatures["NumberOfCategoricalFeatures"](X, y, logger, feat_type).value
        )
        num_numerical = float(
            metafeatures["NumberOfNumericFeatures"](X, y, logger, feat_type).value
        )
        if num_numerical == 0.0:
            return 0.0
        else:
            return num_categorical / num_numerical


# Number of attributes divided by number of samples
@metafeatures.define("DatasetRatio")
class DatasetRatio(MetaFeature):
    def _calculate(self, X, y, logger, feat_type):
        return float(metafeatures["NumberOfFeatures"](X, y, logger).value) / float(
            metafeatures["NumberOfInstances"](X, y, logger).value
        )


@metafeatures.define("LogDatasetRatio", dependency="DatasetRatio")
class LogDatasetRatio(MetaFeature):
    def _calculate(self, X, y, logger, feat_type):
        return np.log(metafeatures.get_value("DatasetRatio"))


@metafeatures.define("InverseDatasetRatio")
class InverseDatasetRatio(MetaFeature):
    def _calculate(self, X, y, logger, feat_type):
        return float(metafeatures["NumberOfInstances"](X, y, logger).value) / float(
            metafeatures["NumberOfFeatures"](X, y, logger).value
        )


@metafeatures.define("LogInverseDatasetRatio", dependency="InverseDatasetRatio")
class LogInverseDatasetRatio(MetaFeature):
    def _calculate(self, X, y, logger, feat_type):
        return np.log(metafeatures.get_value("InverseDatasetRatio"))


@helper_functions.define("ClassOccurences")
class ClassOccurences(HelperFunction):
    def _calculate(self, X, y, logger, feat_type):
        if len(y.shape) == 2:
            occurences = []
            for i in range(y.shape[1]):
                occurences.append(self._calculate(X, y[:, i], logger, feat_type))
            return occurences
        else:
            occurence_dict = defaultdict(float)
            for value in y:
                occurence_dict[value] += 1
            return occurence_dict


@metafeatures.define("ClassProbabilityMin", dependency="ClassOccurences")
class ClassProbabilityMin(MetaFeature):
    def _calculate(self, X, y, logger, feat_type):
        occurences = helper_functions.get_value("ClassOccurences")

        min_value = np.iinfo(np.int64).max
        if len(y.shape) == 2:
            for i in range(y.shape[1]):
                for num_occurences in occurences[i].values():
                    if num_occurences < min_value:
                        min_value = num_occurences
        else:
            for num_occurences in occurences.values():
                if num_occurences < min_value:
                    min_value = num_occurences
        return float(min_value) / float(y.shape[0])


# aka default accuracy
@metafeatures.define("ClassProbabilityMax", dependency="ClassOccurences")
class ClassProbabilityMax(MetaFeature):
    def _calculate(self, X, y, logger, feat_type):
        occurences = helper_functions.get_value("ClassOccurences")
        max_value = -1

        if len(y.shape) == 2:
            for i in range(y.shape[1]):
                for num_occurences in occurences[i].values():
                    if num_occurences > max_value:
                        max_value = num_occurences
        else:
            for num_occurences in occurences.values():
                if num_occurences > max_value:
                    max_value = num_occurences
        return float(max_value) / float(y.shape[0])


@metafeatures.define("ClassProbabilityMean", dependency="ClassOccurences")
class ClassProbabilityMean(MetaFeature):
    def _calculate(self, X, y, logger, feat_type):
        occurence_dict = helper_functions.get_value("ClassOccurences")

        if len(y.shape) == 2:
            occurences = []
            for i in range(y.shape[1]):
                occurences.extend(
                    [occurrence for occurrence in occurence_dict[i].values()]
                )
            occurences = np.array(occurences)
        else:
            occurences = np.array(
                [occurrence for occurrence in occurence_dict.values()], dtype=np.float64
            )
        return (occurences / y.shape[0]).mean()


@metafeatures.define("ClassProbabilitySTD", dependency="ClassOccurences")
class ClassProbabilitySTD(MetaFeature):
    def _calculate(self, X, y, logger, feat_type):
        occurence_dict = helper_functions.get_value("ClassOccurences")

        if len(y.shape) == 2:
            stds = []
            for i in range(y.shape[1]):
                std = np.array(
                    [occurrence for occurrence in occurence_dict[i].values()],
                    dtype=np.float64,
                )
                std = (std / y.shape[0]).std()
                stds.append(std)
            return np.mean(stds)
        else:
            occurences = np.array(
                [occurrence for occurrence in occurence_dict.values()], dtype=np.float64
            )
            return (occurences / y.shape[0]).std()


################################################################################
# Reif, A Comprehensive Dataset for Evaluating Approaches of various Meta-Learning Tasks
# defines these five metafeatures as simple metafeatures, but they could also
#  be the counterpart for the skewness and kurtosis of the numerical features
@helper_functions.define("NumSymbols")
class NumSymbols(HelperFunction):
    def _calculate(self, X, y, logger, feat_type):
        categorical = {
            key: True if value.lower() == "categorical" else False
            for key, value in feat_type.items()
        }
        symbols_per_column = []
        for i in range(X.shape[1]):
            if categorical[X.columns[i] if hasattr(X, "columns") else i]:
                column = X.iloc[:, i] if hasattr(X, "iloc") else X[:, i]
                unique_values = (
                    column.unique() if hasattr(column, "unique") else np.unique(column)
                )
                num_unique = np.sum(pd.notna(unique_values))
                symbols_per_column.append(num_unique)
        return symbols_per_column

    def _calculate_sparse(self, X, y, logger, feat_type):
        categorical = {
            key: True if value.lower() == "categorical" else False
            for key, value in feat_type.items()
        }
        symbols_per_column = []
        new_X = X.tocsc()
        for i in range(new_X.shape[1]):
            if categorical[X.columns[i] if hasattr(X, "columns") else i]:
                unique_values = np.unique(new_X.getcol(i).data)
                num_unique = np.sum(np.isfinite(unique_values))
                symbols_per_column.append(num_unique)
        return symbols_per_column


@metafeatures.define("SymbolsMin", dependency="NumSymbols")
class SymbolsMin(MetaFeature):
    def _calculate(self, X, y, logger, feat_type):
        # The minimum can only be zero if there are no nominal features,
        # otherwise it is at least one
        # TODO: shouldn't this rather be two?
        minimum = None
        for unique in helper_functions.get_value("NumSymbols"):
            if unique > 0 and (minimum is None or unique < minimum):
                minimum = unique
        return minimum if minimum is not None else 0


@metafeatures.define("SymbolsMax", dependency="NumSymbols")
class SymbolsMax(MetaFeature):
    def _calculate(self, X, y, logger, feat_type):
        values = helper_functions.get_value("NumSymbols")
        if len(values) == 0:
            return 0
        return max(max(values), 0)


@metafeatures.define("SymbolsMean", dependency="NumSymbols")
class SymbolsMean(MetaFeature):
    def _calculate(self, X, y, logger, feat_type):
        # TODO: categorical attributes without a symbol don't count towards this
        # measure
        values = [val for val in helper_functions.get_value("NumSymbols") if val > 0]
        mean = np.nanmean(values)
        return mean if np.isfinite(mean) else 0


@metafeatures.define("SymbolsSTD", dependency="NumSymbols")
class SymbolsSTD(MetaFeature):
    def _calculate(self, X, y, logger, feat_type):
        values = [val for val in helper_functions.get_value("NumSymbols") if val > 0]
        std = np.nanstd(values)
        return std if np.isfinite(std) else 0


@metafeatures.define("SymbolsSum", dependency="NumSymbols")
class SymbolsSum(MetaFeature):
    def _calculate(self, X, y, logger, feat_type):
        sum = np.nansum(helper_functions.get_value("NumSymbols"))
        return sum if np.isfinite(sum) else 0


################################################################################
# Statistical meta features
# Only use third and fourth statistical moment because it is common to
# standardize for the other two
# see Engels & Theusinger, 1998 - Using a Data Metric for Preprocessing Advice
# for Data Mining Applications.


@helper_functions.define("Kurtosisses")
class Kurtosisses(HelperFunction):
    def _calculate(self, X, y, logger, feat_type):
        numerical = {
            key: True if value.lower() == "numerical" else False
            for key, value in feat_type.items()
        }
        kurts = []
        for i in range(X.shape[1]):
            if numerical[X.columns[i] if hasattr(X, "columns") else i]:
                if np.isclose(
                    np.var(X.iloc[:, i] if hasattr(X, "iloc") else X[:, i]), 0
                ):
                    kurts.append(0)
                else:
                    kurts.append(
                        scipy.stats.kurtosis(
                            X.iloc[:, i] if hasattr(X, "iloc") else X[:, i]
                        )
                    )
        return kurts

    def _calculate_sparse(self, X, y, logger, feat_type):
        numerical = {
            key: True if value.lower() == "numerical" else False
            for key, value in feat_type.items()
        }
        kurts = []
        X_new = X.tocsc()
        for i in range(X_new.shape[1]):
            if numerical[X.columns[i] if hasattr(X, "columns") else i]:
                start = X_new.indptr[i]
                stop = X_new.indptr[i + 1]
                if np.isclose(np.var(X_new.data[start:stop]), 0):
                    kurts.append(0)
                else:
                    kurts.append(scipy.stats.kurtosis(X_new.data[start:stop]))
        return kurts


@metafeatures.define("KurtosisMin", dependency="Kurtosisses")
class KurtosisMin(MetaFeature):
    def _calculate(self, X, y, logger, feat_type):
        kurts = helper_functions.get_value("Kurtosisses")
        minimum = np.nanmin(kurts) if len(kurts) > 0 else 0
        return minimum if np.isfinite(minimum) else 0


@metafeatures.define("KurtosisMax", dependency="Kurtosisses")
class KurtosisMax(MetaFeature):
    def _calculate(self, X, y, logger, feat_type):
        kurts = helper_functions.get_value("Kurtosisses")
        maximum = np.nanmax(kurts) if len(kurts) > 0 else 0
        return maximum if np.isfinite(maximum) else 0


@metafeatures.define("KurtosisMean", dependency="Kurtosisses")
class KurtosisMean(MetaFeature):
    def _calculate(self, X, y, logger, feat_type):
        kurts = helper_functions.get_value("Kurtosisses")
        mean = np.nanmean(kurts) if len(kurts) > 0 else 0
        return mean if np.isfinite(mean) else 0


@metafeatures.define("KurtosisSTD", dependency="Kurtosisses")
class KurtosisSTD(MetaFeature):
    def _calculate(self, X, y, logger, feat_type):
        kurts = helper_functions.get_value("Kurtosisses")
        std = np.nanstd(kurts) if len(kurts) > 0 else 0
        return std if np.isfinite(std) else 0


@helper_functions.define("Skewnesses")
class Skewnesses(HelperFunction):
    def _calculate(self, X, y, logger, feat_type):
        numerical = {
            key: True if value.lower() == "numerical" else False
            for key, value in feat_type.items()
        }
        skews = []
        for i in range(X.shape[1]):
            if numerical[X.columns[i] if hasattr(X, "columns") else i]:
                if np.isclose(
                    np.var(X.iloc[:, i] if hasattr(X, "iloc") else X[:, i]), 0
                ):
                    skews.append(0)
                else:
                    skews.append(
                        scipy.stats.skew(
                            X.iloc[:, i] if hasattr(X, "iloc") else X[:, i]
                        )
                    )
        return skews

    def _calculate_sparse(self, X, y, logger, feat_type):
        numerical = {
            key: True if value.lower() == "numerical" else False
            for key, value in feat_type.items()
        }
        skews = []
        X_new = X.tocsc()
        for i in range(X_new.shape[1]):
            if numerical[X.columns[i] if hasattr(X, "columns") else i]:
                start = X_new.indptr[i]
                stop = X_new.indptr[i + 1]
                if np.isclose(np.var(X_new.data[start:stop]), 0):
                    skews.append(0)
                else:
                    skews.append(scipy.stats.skew(X_new.data[start:stop]))
        return skews


@metafeatures.define("SkewnessMin", dependency="Skewnesses")
class SkewnessMin(MetaFeature):
    def _calculate(self, X, y, logger, feat_type):
        skews = helper_functions.get_value("Skewnesses")
        minimum = np.nanmin(skews) if len(skews) > 0 else 0
        return minimum if np.isfinite(minimum) else 0


@metafeatures.define("SkewnessMax", dependency="Skewnesses")
class SkewnessMax(MetaFeature):
    def _calculate(self, X, y, logger, feat_type):
        skews = helper_functions.get_value("Skewnesses")
        maximum = np.nanmax(skews) if len(skews) > 0 else 0
        return maximum if np.isfinite(maximum) else 0


@metafeatures.define("SkewnessMean", dependency="Skewnesses")
class SkewnessMean(MetaFeature):
    def _calculate(self, X, y, logger, feat_type):
        skews = helper_functions.get_value("Skewnesses")
        mean = np.nanmean(skews) if len(skews) > 0 else 0
        return mean if np.isfinite(mean) else 0


@metafeatures.define("SkewnessSTD", dependency="Skewnesses")
class SkewnessSTD(MetaFeature):
    def _calculate(self, X, y, logger, feat_type):
        skews = helper_functions.get_value("Skewnesses")
        std = np.nanstd(skews) if len(skews) > 0 else 0
        return std if np.isfinite(std) else 0


"""
@metafeatures.define("cancor1")
def cancor1(X, y):
    pass

@metafeatures.define("cancor2")
def cancor2(X, y):
    pass
"""


################################################################################
# Information-theoretic metafeatures
@metafeatures.define("ClassEntropy")
class ClassEntropy(MetaFeature):
    def _calculate(self, X, y, logger, feat_type):
        labels = 1 if len(y.shape) == 1 else y.shape[1]

        entropies = []
        for i in range(labels):
            occurence_dict = defaultdict(float)
            for value in y if labels == 1 else y[:, i]:
                occurence_dict[value] += 1
            entropies.append(
                scipy.stats.entropy(
                    [occurence_dict[key] for key in occurence_dict], base=2
                )
            )

        return np.mean(entropies)


"""
#@metafeatures.define("normalized_class_entropy")

#@metafeatures.define("attribute_entropy")

#@metafeatures.define("normalized_attribute_entropy")

#@metafeatures.define("joint_entropy")

#@metafeatures.define("mutual_information")

#@metafeatures.define("noise-signal-ratio")

#@metafeatures.define("signal-noise-ratio")

#@metafeatures.define("equivalent_number_of_attributes")

#@metafeatures.define("conditional_entropy")

#@metafeatures.define("average_attribute_entropy")
"""


################################################################################
# Landmarking features, computed with cross validation
# These should be invoked with the same transformations of X and y with which
# sklearn will be called later on


# from Pfahringer 2000
# Linear discriminant learner
@metafeatures.define("LandmarkLDA")
class LandmarkLDA(MetaFeature):
    def _calculate(self, X, y, logger, feat_type):
        import sklearn.discriminant_analysis

        if type(y) in ("binary", "multiclass"):
            kf = sklearn.model_selection.StratifiedKFold(n_splits=5)
        else:
            kf = sklearn.model_selection.KFold(n_splits=5)

        accuracy = 0.0
        try:
            for train, test in kf.split(X, y):
                lda = sklearn.discriminant_analysis.LinearDiscriminantAnalysis()

                if len(y.shape) == 1 or y.shape[1] == 1:
                    lda.fit(
                        X.iloc[train] if hasattr(X, "iloc") else X[train],
                        y.iloc[train] if hasattr(y, "iloc") else y[train],
                    )
                else:
                    lda = OneVsRestClassifier(lda)
                    lda.fit(
                        X.iloc[train] if hasattr(X, "iloc") else X[train],
                        y.iloc[train] if hasattr(y, "iloc") else y[train],
                    )

                predictions = lda.predict(
                    X.iloc[test] if hasattr(X, "iloc") else X[test],
                )
                accuracy += sklearn.metrics.accuracy_score(
                    predictions,
                    y.iloc[test] if hasattr(y, "iloc") else y[test],
                )
            return accuracy / 5
        except scipy.linalg.LinAlgError as e:
            self.logger.warning("LDA failed: %s Returned 0 instead!" % e)
            return np.NaN
        except ValueError as e:
            self.logger.warning("LDA failed: %s Returned 0 instead!" % e)
            return np.NaN

    def _calculate_sparse(self, X, y, logger, feat_type):
        return np.NaN


# Naive Bayes
@metafeatures.define("LandmarkNaiveBayes")
class LandmarkNaiveBayes(MetaFeature):
    def _calculate(self, X, y, logger, feat_type):
        import sklearn.naive_bayes

        if type(y) in ("binary", "multiclass"):
            kf = sklearn.model_selection.StratifiedKFold(n_splits=5)
        else:
            kf = sklearn.model_selection.KFold(n_splits=5)

        accuracy = 0.0
        for train, test in kf.split(X, y):
            nb = sklearn.naive_bayes.GaussianNB()

            if len(y.shape) == 1 or y.shape[1] == 1:
                nb.fit(
                    X.iloc[train] if hasattr(X, "iloc") else X[train],
                    y.iloc[train] if hasattr(y, "iloc") else y[train],
                )
            else:
                nb = OneVsRestClassifier(nb)
                nb.fit(
                    X.iloc[train] if hasattr(X, "iloc") else X[train],
                    y.iloc[train] if hasattr(y, "iloc") else y[train],
                )

            predictions = nb.predict(
                X.iloc[test] if hasattr(X, "iloc") else X[test],
            )
            accuracy += sklearn.metrics.accuracy_score(
                predictions,
                y.iloc[test] if hasattr(y, "iloc") else y[test],
            )
        return accuracy / 5

    def _calculate_sparse(self, X, y, logger, feat_type):
        return np.NaN


# Cart learner instead of C5.0
@metafeatures.define("LandmarkDecisionTree")
class LandmarkDecisionTree(MetaFeature):
    def _calculate(self, X, y, logger, feat_type):
        import sklearn.tree

        if type(y) in ("binary", "multiclass"):
            kf = sklearn.model_selection.StratifiedKFold(n_splits=5)
        else:
            kf = sklearn.model_selection.KFold(n_splits=5)

        accuracy = 0.0
        for train, test in kf.split(X, y):
            random_state = sklearn.utils.check_random_state(42)
            tree = sklearn.tree.DecisionTreeClassifier(random_state=random_state)

            if len(y.shape) == 1 or y.shape[1] == 1:
                tree.fit(
                    X.iloc[train] if hasattr(X, "iloc") else X[train],
                    y.iloc[train] if hasattr(y, "iloc") else y[train],
                )
            else:
                tree = OneVsRestClassifier(tree)
                tree.fit(
                    X.iloc[train] if hasattr(X, "iloc") else X[train],
                    y.iloc[train] if hasattr(y, "iloc") else y[train],
                )

            predictions = tree.predict(
                X.iloc[test] if hasattr(X, "iloc") else X[test],
            )
            accuracy += sklearn.metrics.accuracy_score(
                predictions,
                y.iloc[test] if hasattr(y, "iloc") else y[test],
            )
        return accuracy / 5

    def _calculate_sparse(self, X, y, logger, feat_type):
        return np.NaN


"""If there is a dataset which has OneHotEncoded features it can happend that
a node learner splits at one of the attribute encodings. This should be fine
as the dataset is later on used encoded."""


# TODO: use the same tree, this has then to be computed only once and hence
#  saves a lot of time...
@metafeatures.define("LandmarkDecisionNodeLearner")
class LandmarkDecisionNodeLearner(MetaFeature):
    def _calculate(self, X, y, logger, feat_type):
        import sklearn.tree

        if type(y) in ("binary", "multiclass"):
            kf = sklearn.model_selection.StratifiedKFold(n_splits=5)
        else:
            kf = sklearn.model_selection.KFold(n_splits=5)

        accuracy = 0.0
        for train, test in kf.split(X, y):
            random_state = sklearn.utils.check_random_state(42)
            node = sklearn.tree.DecisionTreeClassifier(
                criterion="entropy",
                max_depth=1,
                random_state=random_state,
                min_samples_split=2,
                min_samples_leaf=1,
                max_features=None,
            )
            if len(y.shape) == 1 or y.shape[1] == 1:
                node.fit(
                    X.iloc[train] if hasattr(X, "iloc") else X[train],
                    y.iloc[train] if hasattr(y, "iloc") else y[train],
                )
            else:
                node = OneVsRestClassifier(node)
                node.fit(
                    X.iloc[train] if hasattr(X, "iloc") else X[train],
                    y.iloc[train] if hasattr(y, "iloc") else y[train],
                )
            predictions = node.predict(
                X.iloc[test] if hasattr(X, "iloc") else X[test],
            )
            accuracy += sklearn.metrics.accuracy_score(
                predictions,
                y.iloc[test] if hasattr(y, "iloc") else y[test],
            )
        return accuracy / 5

    def _calculate_sparse(self, X, y, logger, feat_type):
        return np.NaN


@metafeatures.define("LandmarkRandomNodeLearner")
class LandmarkRandomNodeLearner(MetaFeature):
    def _calculate(self, X, y, logger, feat_type):
        import sklearn.tree

        if type(y) in ("binary", "multiclass"):
            kf = sklearn.model_selection.StratifiedKFold(n_splits=5)
        else:
            kf = sklearn.model_selection.KFold(n_splits=5)
        accuracy = 0.0

        for train, test in kf.split(X, y):
            random_state = sklearn.utils.check_random_state(42)
            node = sklearn.tree.DecisionTreeClassifier(
                criterion="entropy",
                max_depth=1,
                random_state=random_state,
                min_samples_split=2,
                min_samples_leaf=1,
                max_features=1,
            )
            node.fit(
                X.iloc[train] if hasattr(X, "iloc") else X[train],
                y.iloc[train] if hasattr(y, "iloc") else y[train],
            )
            predictions = node.predict(
                X.iloc[test] if hasattr(X, "iloc") else X[test],
            )
            accuracy += sklearn.metrics.accuracy_score(
                predictions,
                y.iloc[test] if hasattr(y, "iloc") else y[test],
            )
        return accuracy / 5

    def _calculate_sparse(self, X, y, logger, feat_type):
        return np.NaN


"""
This is wrong...
@metafeatures.define("landmark_worst_node_learner")
def landmark_worst_node_learner(X, y):
    # TODO: this takes more than 10 minutes on some datasets (eg mfeat-pixels)
    # which has 240*6 = 1440 discrete attributes...
    # TODO: calculate information gain instead of using the worst test result
    import sklearn.tree
    performances = []
    for attribute_idx in range(X.shape[1]):
        kf = sklearn.model_selection.StratifiedKFold(y, n_folds=5)
        accuracy = 0.
        for train, test in kf:
            node = sklearn.tree.DecisionTreeClassifier(criterion="entropy",
                max_features=None, max_depth=1, min_samples_split=1,
                min_samples_leaf=1)
            node.fit(X[train][:,attribute_idx].reshape((-1, 1)), y[train],
                     check_input=False)
            predictions = node.predict(X[test][:,attribute_idx].reshape((-1, 1)))
            accuracy += sklearn.metrics.accuracy_score(predictions, y[test])
        performances.append(1 - (accuracy / 10))
    return max(performances)
"""


# Replace the Elite 1NN with a normal 1NN, this slightly changes the
# intuition behind this landmark, but Elite 1NN is used nowhere else...
@metafeatures.define("Landmark1NN")
class Landmark1NN(MetaFeature):
    def _calculate(self, X, y, logger, feat_type):
        import sklearn.neighbors

        if type(y) in ("binary", "multiclass"):
            kf = sklearn.model_selection.StratifiedKFold(n_splits=5)
        else:
            kf = sklearn.model_selection.KFold(n_splits=5)

        accuracy = 0.0
        for train, test in kf.split(X, y):
            kNN = sklearn.neighbors.KNeighborsClassifier(n_neighbors=1)
            if len(y.shape) == 1 or y.shape[1] == 1:
                kNN.fit(
                    X.iloc[train] if hasattr(X, "iloc") else X[train],
                    y.iloc[train] if hasattr(y, "iloc") else y[train],
                )
            else:
                kNN = OneVsRestClassifier(kNN)
                kNN.fit(
                    X.iloc[train] if hasattr(X, "iloc") else X[train],
                    y.iloc[train] if hasattr(y, "iloc") else y[train],
                )
            predictions = kNN.predict(
                X.iloc[test] if hasattr(X, "iloc") else X[test],
            )
            accuracy += sklearn.metrics.accuracy_score(
                predictions,
                y.iloc[test] if hasattr(y, "iloc") else y[test],
            )
        return accuracy / 5


################################################################################
# Bardenet 2013 - Collaborative Hyperparameter Tuning
# K number of classes ("number_of_classes")
# log(d), log(number of attributes)
# log(n/d), log(number of training instances/number of attributes)
# p, how many principal components to keep in order to retain 95% of the
#     dataset variance
# skewness of a dataset projected onto one principal component...
# kurtosis of a dataset projected onto one principal component
@helper_functions.define("PCA")
class PCA(HelperFunction):
    def _calculate(self, X, y, logger, feat_type):
        import sklearn.decomposition

        pca = sklearn.decomposition.PCA(copy=True)
        rs = np.random.RandomState(42)
        indices = np.arange(X.shape[0])
        for i in range(10):
            try:
                rs.shuffle(indices)
                pca.fit(
                    X.iloc[indices] if hasattr(X, "iloc") else X[indices],
                )
                return pca
            except LinAlgError:
                pass
        self.logger.warning("Failed to compute a Principle Component Analysis")
        return None

    def _calculate_sparse(self, X, y, logger, feat_type):
        import sklearn.decomposition

        rs = np.random.RandomState(42)
        indices = np.arange(X.shape[0])
        # This is expensive, but necessary with scikit-learn 0.15
        Xt = X.astype(np.float64)
        for i in range(10):
            try:
                rs.shuffle(indices)
                truncated_svd = sklearn.decomposition.TruncatedSVD(
                    n_components=X.shape[1] - 1, random_state=i, algorithm="randomized"
                )
                truncated_svd.fit(Xt[indices])
                return truncated_svd
            except LinAlgError:
                pass
        self.logger.warning("Failed to compute a Truncated SVD")
        return None


# Maybe define some more...
@metafeatures.define("PCAFractionOfComponentsFor95PercentVariance", dependency="PCA")
class PCAFractionOfComponentsFor95PercentVariance(MetaFeature):
    def _calculate(self, X, y, logger, feat_type):
        pca_ = helper_functions.get_value("PCA")
        if pca_ is None:
            return np.NaN
        sum_ = 0.0
        idx = 0
        while sum_ < 0.95 and idx < len(pca_.explained_variance_ratio_):
            sum_ += pca_.explained_variance_ratio_[idx]
            idx += 1
        return float(idx) / float(X.shape[1])


# Kurtosis of first PC
@metafeatures.define("PCAKurtosisFirstPC", dependency="PCA")
class PCAKurtosisFirstPC(MetaFeature):
    def _calculate(self, X, y, logger, feat_type):
        pca_ = helper_functions.get_value("PCA")
        if pca_ is None:
            return np.NaN
        components = pca_.components_
        pca_.components_ = components[:1]
        transformed = pca_.transform(X)
        pca_.components_ = components

        kurtosis = scipy.stats.kurtosis(transformed)
        return kurtosis[0]


# Skewness of first PC
@metafeatures.define("PCASkewnessFirstPC", dependency="PCA")
class PCASkewnessFirstPC(MetaFeature):
    def _calculate(self, X, y, logger, feat_type):
        pca_ = helper_functions.get_value("PCA")
        if pca_ is None:
            return np.NaN
        components = pca_.components_
        pca_.components_ = components[:1]
        transformed = pca_.transform(X)
        pca_.components_ = components

        skewness = scipy.stats.skew(transformed)
        return skewness[0]


def calculate_all_metafeatures_encoded_labels(
    X, y, feat_type, dataset_name, logger, calculate=None, dont_calculate=None
):
    """
    Calculate only metafeatures for which a 1HotEncoded feature matrix is necessery.
    """

    calculate = set()
    calculate.update(npy_metafeatures)

    return calculate_all_metafeatures(
        X,
        y,
        feat_type,
        dataset_name,
        calculate=calculate,
        dont_calculate=dont_calculate,
        logger=logger,
    )


def calculate_all_metafeatures_with_labels(
    X, y, feat_type, dataset_name, logger, calculate=None, dont_calculate=None
):
    if dont_calculate is None:
        dont_calculate = set()
    else:
        dont_calculate = copy.deepcopy(dont_calculate)
    dont_calculate.update(npy_metafeatures)
    return calculate_all_metafeatures(
        X,
        y,
        feat_type,
        dataset_name,
        calculate=calculate,
        dont_calculate=dont_calculate,
        logger=logger,
    )


def calculate_all_metafeatures(
    X,
    y,
    feat_type,
    dataset_name,
    logger,
    calculate=None,
    dont_calculate=None,
    densify_threshold=1000,
):

    """Calculate all metafeatures."""
    helper_functions.clear()
    metafeatures.clear()
    mf_ = dict()

    visited = set()
    to_visit = deque()
    to_visit.extend(metafeatures)

    X_transformed = None
    y_transformed = None

    # TODO calculate the numpy metafeatures after all others to consume less
    # memory
    while len(to_visit) > 0:
        name = to_visit.pop()
        if calculate is not None and name not in calculate:
            continue
        if dont_calculate is not None and name in dont_calculate:
            continue

        if name in npy_metafeatures:
            if X_transformed is None:
                # TODO make sure this is done as efficient as possible (no copy for
                # sparse matrices because of wrong sparse format)
                sparse = scipy.sparse.issparse(X)

                # TODO make this more cohesive to the overall structure (quick bug fix)
                if isinstance(X, pd.DataFrame):
                    for key in X.select_dtypes(include="string").columns:
                        feat_type[key] = "string"

                DPP = FeatTypeSplit(
                    # The difference between feat_type and categorical, is that
                    # categorical has True/False instead of categorical/numerical
                    feat_type=feat_type,
                    force_sparse_output=True,
                )
                X_transformed = DPP.fit_transform(X)
                feat_type_transformed = {
                    i: "numerical" for i in range(X_transformed.shape[1])
                }

                # Densify the transformed matrix
                if not sparse and scipy.sparse.issparse(X_transformed):
                    bytes_per_float = X_transformed.dtype.itemsize
                    num_elements = X_transformed.shape[0] * X_transformed.shape[1]
                    megabytes_required = num_elements * bytes_per_float / 1000 / 1000
                    if megabytes_required < densify_threshold:
                        X_transformed = X_transformed.todense()

                # This is not only important for datasets which are somehow
                # sorted in a strange way, but also prevents lda from failing in
                # some cases.
                # Because this is advanced indexing, a copy of the data is returned!!!
                X_transformed = check_array(
                    X_transformed, force_all_finite=True, accept_sparse="csr"
                )
                indices = np.arange(X_transformed.shape[0])

                rs = np.random.RandomState(42)
                rs.shuffle(indices)

                # TODO Shuffle inplace
                X_transformed = X_transformed[indices]
                y_transformed = y[indices]

            X_ = X_transformed
            y_ = y_transformed
            feat_type_ = feat_type_transformed
        else:
            X_ = X
            y_ = y
            feat_type_ = feat_type

        dependency = metafeatures.get_dependency(name)
        if dependency is not None:
            is_metafeature = dependency in metafeatures
            is_helper_function = dependency in helper_functions

            if is_metafeature and is_helper_function:
                raise NotImplementedError()
            elif not is_metafeature and not is_helper_function:
                raise ValueError(dependency)
            elif is_metafeature and not metafeatures.is_calculated(dependency):
                to_visit.appendleft(name)
                continue
            elif is_helper_function and not helper_functions.is_calculated(dependency):
                logger.info("%s: Going to calculate: %s", dataset_name, dependency)
                value = helper_functions[dependency](
                    X_, y_, feat_type=feat_type_, logger=logger
                )
                helper_functions.set_value(dependency, value)
                mf_[dependency] = value

        logger.info("%s: Going to calculate: %s", dataset_name, name)

        value = metafeatures[name](X_, y_, logger, feat_type_)
        metafeatures.set_value(name, value)
        mf_[name] = value
        visited.add(name)

    mf_ = DatasetMetafeatures(dataset_name, mf_)
    return mf_


npy_metafeatures = set(
    [
        "LandmarkLDA",
        "LandmarkNaiveBayes",
        "LandmarkDecisionTree",
        "LandmarkDecisionNodeLearner",
        "LandmarkRandomNodeLearner",
        "LandmarkWorstNodeLearner",
        "Landmark1NN",
        "PCAFractionOfComponentsFor95PercentVariance",
        "PCAKurtosisFirstPC",
        "PCASkewnessFirstPC",
        "Skewnesses",
        "SkewnessMin",
        "SkewnessMax",
        "SkewnessMean",
        "SkewnessSTD",
        "Kurtosisses",
        "KurtosisMin",
        "KurtosisMax",
        "KurtosisMean",
        "KurtosisSTD",
    ]
)

subsets = dict()
# All implemented metafeatures
subsets["all"] = set(metafeatures.functions.keys())

# Metafeatures used by Pfahringer et al. (2000) in the first experiment
subsets["pfahringer_2000_experiment1"] = set(
    [
        "number_of_features",
        "number_of_numeric_features",
        "number_of_categorical_features",
        "number_of_classes",
        "class_probability_max",
        "landmark_lda",
        "landmark_naive_bayes",
        "landmark_decision_tree",
    ]
)

# Metafeatures used by Pfahringer et al. (2000) in the second experiment
# worst node learner not implemented yet
"""
pfahringer_2000_experiment2 = set(["landmark_decision_node_learner",
                                   "landmark_random_node_learner",
                                   "landmark_worst_node_learner",
                                   "landmark_1NN"])
"""

# Metafeatures used by Yogatama and Mann (2014)
subsets["yogotama_2014"] = set(
    ["log_number_of_features", "log_number_of_instances", "number_of_classes"]
)

# Metafeatures used by Bardenet et al. (2013) for the AdaBoost.MH experiment
subsets["bardenet_2013_boost"] = set(
    [
        "number_of_classes",
        "log_number_of_features",
        "log_inverse_dataset_ratio",
        "pca_95percent",
    ]
)

# Metafeatures used by Bardenet et al. (2013) for the Neural Net experiment
subsets["bardenet_2013_nn"] = set(
    [
        "number_of_classes",
        "log_number_of_features",
        "log_inverse_dataset_ratio",
        "pca_kurtosis_first_pc",
        "pca_skewness_first_pc",
    ]
)



from typing import List, Optional, Union, no_type_check

import importlib
from distutils.version import LooseVersion

import pkg_resources

from autosklearn.util import RE_PATTERN


def verify_packages(packages: Optional[Union[str, List[str]]]) -> None:
    if not packages:
        return
    if isinstance(packages, str):
        packages = packages.splitlines()

    for package in packages:
        if not package:
            continue

        # Ignore comments
        if package.startswith("#"):
            continue

        match = RE_PATTERN.match(package)
        if match:
            name = match.group("name")
            operation = match.group("operation1")
            version = match.group("version1")
            _verify_package(name, operation, version)
        else:
            raise ValueError("Unable to read requirement: %s" % package)


# Module has no attribute __version__ wa
@no_type_check
def _verify_package(name: str, operation: Optional[str], version: str) -> None:
    try:
        module_dist = pkg_resources.get_distribution(name)
        installed_version = LooseVersion(module_dist.version)
    except pkg_resources.DistributionNotFound:
        try:
            module = importlib.import_module(name)  # type: ignore
            installed_version = LooseVersion(module.__version__)
        except ImportError:
            raise MissingPackageError(name)

    if not operation:
        return

    required_version = LooseVersion(version)

    if operation == "==":
        check = required_version == installed_version
    elif operation == ">":
        check = installed_version > required_version
    elif operation == "<":
        check = installed_version < required_version
    elif operation == ">=":
        check = (
            installed_version > required_version
            or installed_version == required_version
        )
    else:
        raise NotImplementedError("operation '%s' is not supported" % operation)
    if not check:
        raise IncorrectPackageVersionError(
            name, installed_version, operation, required_version
        )


class MissingPackageError(Exception):
    error_message = "Mandatory package '{name}' not found!"

    def __init__(self, package_name: str):
        self.package_name = package_name
        super(MissingPackageError, self).__init__(
            self.error_message.format(name=package_name)
        )


class IncorrectPackageVersionError(Exception):
    error_message = (
        "found '{name}' version {installed_version} but requires {name} version "
        "{operation}{required_version}"
    )

    def __init__(
        self,
        package_name: str,
        installed_version: Union[str, LooseVersion],
        operation: Optional[str],
        required_version: Union[str, LooseVersion],
    ):
        self.package_name = package_name
        self.installed_version = installed_version
        self.operation = operation
        self.required_version = required_version
        message = self.error_message.format(
            name=package_name,
            installed_version=installed_version,
            operation=operation,
            required_version=required_version,
        )
        super(IncorrectPackageVersionError, self).__init__(message)
